/* This program generates a dynagraph data file.  Output goes to stdout.

   Run the program, redirect stdout to a file, then display the image
   in dynagraph with the command:
       dataplot("filename");

   The image shown is modeled after the "Umbilic Torus NC" sculpture
   of Helaman Ferguson.

   See http://www.cs.berkeley.edu/~sequin/SCULPTS/ferguson.html for
   a photo of the sculpture.

   See http://www.access.digex.net/~helamanf/ for more info and images.

   -- RR, August 1999
*/

#include <stdio.h>
#include <math.h>

#define PI	3.14159265358979323846

int main(void)
{
	float r = 0.5;
	float R = 1.1;
	float twists = +2.0/3;

	float x, y, z, X, Z, L;
	float a, s, t;
	int i, j;
	int m = 40;
	int n = 9;

	printf("# Display this datafile in dynagraph by:\n");
	printf("#  dataplot(`filename`);\n");
	printf("\n");
	printf("TITLE 'Umbilic Torus NC (after Helaman Ferguson)'\n\n");

	printf("SURF %d %d\n", m, n);
	for (i=0; i<m; i++)  {
		t = i*2.0*PI/(m-1);		/* longitudinal angle */
		a = t*twists/2.0;		/* slant of the Mobius "ribs" */
		for (j=0; j<n; j++) {
			s = (float)j/(n-1);	/* s goes from 0 to 1 */

			/* [X(s), Y(s)] with s=0..1 is a straight line segment
			   which connects the two points (cos(a), sin(a)) and
			   (cos(a+120), sin(a+120)) of the unit circle in
			   the XZ plane.
			*/
			X = cos(a)-sqrt(3)*s*cos(a-PI/6);
			Z = sin(a)-sqrt(3)*s*sin(a-PI/6);

			/* We compute the distance D = sqrt(X^2+Z^2) of
			   each point on the line segment from the origin,
			   then define L as D^b, where b is a suitably chosen
			   exponent (we take b=1/3 below)
			*/
			L = pow(sqrt(X*X+Z*Z), 1.0/3);

			/* [L*X(s), L*Y(s)], s=0..1 is a curve which
			   connects the two points (cos(a), sin(a)) and
			   (cos(a+120), sin(a+120)) of the unit circle.
			   It is like the line segment described above, but
			   it is "caved in" a bit, passing closer to the
			   origin than the strain line.  The exponent b
			   described before controls the amount of the 
			   caving in.

			   We magnify the curve by a constant factor of r
			   then translate it in the x direction by R, and
			   rotate in the xy plane by angle t.  Then compute
			   the resulting Cartesian coordiantes
			*/
			x = (R + r*L*X)*cos(t);
			y = (R + r*L*X)*sin(t);
			z = r*L*Z;
			printf("%g %g %g\n", x, y, z);
		}
	}

	/* repeat everything, but advance a by 120 degrees */

	printf("SURF %d %d\n", m, n);
	for (i=0; i<m; i++)  {
		t = i*2.0*PI/(m-1);
		a = t*twists/2.0 + 2*PI/3;	/* advance a by 120 degrees */
		for (j=0; j<n; j++) {
			s = (float)j/(n-1);
			X = cos(a)-sqrt(3)*s*cos(a-PI/6);
			Z = sin(a)-sqrt(3)*s*sin(a-PI/6);
			L = pow(sqrt(X*X+Z*Z), 1.0/3);
			x = (R + r*L*X)*cos(t);
			y = (R + r*L*X)*sin(t);
			z = r*L*Z;
			printf("%g %g %g\n", x, y, z);
		}
	}

	/* repeat everything, but advance a by another 120 degrees */

	printf("SURF %d %d\n", m, n);
	for (i=0; i<m; i++)  {
		t = i*2.0*PI/(m-1);
		a = t*twists/2.0 + 4*PI/3;	/* advance a by 120 degrees */
		for (j=0; j<n; j++) {
			s = (float)j/(n-1);
			X = cos(a)-sqrt(3)*s*cos(a-PI/6);
			Z = sin(a)-sqrt(3)*s*sin(a-PI/6);
			L = pow(sqrt(X*X+Z*Z), 1.0/3);
			x = (R + r*L*X)*cos(t);
			y = (R + r*L*X)*sin(t);
			z = r*L*Z;
			printf("%g %g %g\n", x, y, z);
		}
	}

	return 0;
}
