#include <float.h>	/* for DBL_MAX */
#include <stdlib.h>
#include "dynagraph.h"
#include "parse.h"

char *copy_string(char *s)
{
	char *t;

	if (s==NULL)
		return NULL;
	t = malloc(strlen(s)+1);
	strcpy(t, s);
	return t;
}

void free_expr(Expr *e)
{
	if (e==NULL)
		return;
	testfree(e->lval);
	testfree(e->rval);
	free(e);
}

static Expr *copy_expr(Expr *e)
{
	Expr *t;

	if (e==NULL)
		return NULL;
	t = malloc(sizeof(Expr));

	t->lval = copy_string(e->lval);
	t->rval = copy_string(e->rval);
	t->fp = NULL;

	return t;
}

static void free_range(Range *s)
{
	if (s==NULL)
		return;
	testfree(s->id);
	free_expr(s->lo);
	free_expr(s->hi);
	free(s);
}

Range *copy_range(Range *r)
{
	Range *t;

	if (r==NULL)
		return NULL;

	t = malloc(sizeof(Range));

	t->id = copy_string(r->id);
	t->lo = copy_expr(r->lo);
	t->hi = copy_expr(r->hi);

	return t;
}

static void free_expr3(Expr3 *e3)
{
	if (e3==NULL)
		return;
	free_expr(e3->u);
	free_expr(e3->v);
	free_expr(e3->w);
	free(e3);
}

static Expr3 *copy_expr3(Expr3 *e3)
{
	Expr3 *t;

	if (e3==NULL)
		return NULL;
	t = malloc(sizeof(Expr3));
	t->u = copy_expr(e3->u);
	t->v = copy_expr(e3->v);
	t->w = copy_expr(e3->w);

	return t;
}

static void free_polygon_vertices(PolygonVertex *v)
{
	PolygonVertex *p;

	while (v!=NULL) {
		p = v->fp;
		free_expr3(v->v);
		free(v);
		v = p;
	}
}

static PolygonVertex *copy_polygon_vertices(PolygonVertex *v)
{
	PolygonVertex *head = NULL;
	PolygonVertex *tail = NULL;

	while (v!=NULL) {
		PolygonVertex *p = malloc(sizeof(PolygonVertex));
		p->v = copy_expr3(v->v);
		p->numeric.x = v->numeric.x;
		p->numeric.y = v->numeric.y;
		p->numeric.z = v->numeric.z;
		p->fp = NULL;
		if (head==NULL)
			head = tail = p;
		else {
			tail->fp = p;
			tail = p;
		}
		v = v->fp;
	}

	return head;
}

Fontdesc *create_fontdesc(void)
{
	Fontdesc *f;

	f = malloc(sizeof(Fontdesc));
	f->xfontinfo = NULL;
	f->glbase = 0;
	f->fp = NULL;

	return f;
}

static void free_fontdesc(Fontdesc *s)
{
	if (s==NULL)
		return;
	glDeleteLists(s->glbase, s->nchars);
	free(s);
}

TextObj *create_textobject(void)
{
	TextObj *t = malloc(sizeof(TextObj));
	t->text = NULL;
	t->xspec = NULL;
	t->yspec = NULL;
	t->zspec = NULL;
	t->fp = NULL;
	return t;
}

static void free_textobject_list(TextObj *t)
{
	TextObj *tnext;

	while (t!=NULL) {
		tnext = t->fp;
		testfree(t->text);
		free_expr(t->xspec);
		free_expr(t->yspec);
		free_expr(t->zspec);
		free(t);
		t = tnext;
	}
}

static TextObj *copy_textobject_list(TextObj *t)
{
	TextObj *s, *sprev;

	s = sprev = NULL;

	while (t!=NULL) {
		s = create_textobject();
		s->text = copy_string(t->text);
		s->xspec = copy_expr(t->xspec);
		s->yspec = copy_expr(t->yspec);
		s->zspec = copy_expr(t->zspec);
		s->x = t->x;
		s->y = t->y;
		s->z = t->z;
		s->fp = sprev;
		sprev = s;
		t = t->fp;
	}
	return s;
}

Point *create_curvepoints(int m)
{
	Point *curvepoints;

	curvepoints = malloc(m*sizeof(Point));
	return curvepoints;
}

static void free_curvepoints(Point *curvepoints)
{
	testfree(curvepoints);
}

static Point *copy_curvepoints(Point *curvepoints, int m)
{
	Point *t;
	int i;

	if (curvepoints==NULL)
		return NULL;

	t = malloc(m*sizeof(Point));

	for (i=0; i<m; i++) {
		t[i].x = curvepoints[i].x;
		t[i].y = curvepoints[i].y;
		t[i].z = curvepoints[i].z;
	}

	return t;
}

Point **create_mesh(int m, int n)
{
	Point **mesh;
	int i;

	mesh = malloc(m*sizeof(Point *));
	mesh[0] = malloc(m*n*sizeof(Point));
	for (i=1; i<m; i++)
		mesh[i] = mesh[i-1] + n;
	return mesh;
}

void free_mesh(Point **mesh)
{
	if (mesh!=NULL)
		testfree(mesh[0]);
	testfree(mesh);
}

static Point **copy_mesh(Point **mesh, int m, int n)
{
	Point **t;
	int i, j;

	if (mesh==NULL)
		return NULL;
	t = create_mesh(m, n);

	for (i=0; i<m; i++)
		for (j=0; j<n; j++) {
			t[i][j].x = mesh[i][j].x;
			t[i][j].y = mesh[i][j].y;
			t[i][j].z = mesh[i][j].z;
		}

	return t;
}

/* creates a Plot3Dattribs structure; initializes fields with default values */ 
Plot3Dattribs *create_plot3d_attribs_structure(void)
{
	Plot3Dattribs *a;

	a = malloc(sizeof(Plot3Dattribs));
	
	a->id = NULL;

	a->xrange	= NULL;
	a->yrange	= NULL;

	a->xgrid	= NULL;
	a->ygrid	= NULL;

	a->m		= DEFAULT_GRIDSIZE;
	a->n		= DEFAULT_GRIDSIZE;

	a->num_points	= NULL;
	a->num_contours	= NULL;
	a->contours	= DEFAULT_CONTOURS;
	a->tube_points	= NULL;
	a->tube_radius 	= NULL;

	a->idx_overall	= 0;
	a->idx_objects	= 0;
	a->idx_bbox	= 0;
	a->idx_title	= 0;

	a->color	= NULL;
	a->rgbcolors[0]	= -1;
	a->rgbcolors[3]	= ALPHA;

	a->xlabel	= NULL;
	a->ylabel	= NULL;
	a->zlabel	= NULL;

	a->title	= NULL;

	a->view_longitude	= NULL;
	a->view_colatitude	= NULL;
	a->longitude	= DEFAULT_VIEW_LONGITUDE;
	a->colatitude	= DEFAULT_VIEW_COLATITUDE;

	a->axes		= NOAXES;
	a->gridstyle	= XY_GRIDLINES;
	a->scaling	= CONSTRAINED;
	a->shading	= XYZSHADING;
	a->style	= PATCHNOGRID;
	a->coords	= CARTESIAN;

	a->font		= DEFAULT_FONTFAMILY
			+ DEFAULT_FONTSTYLE
			+ DEFAULT_FONTSIZE;
	a->axesfont	= DEFAULT_AXESFONTFAMILY
			+ DEFAULT_AXESFONTSTYLE
			+ DEFAULT_AXESFONTSIZE;
	a->labelfont	= DEFAULT_LABELFONTFAMILY
			+ DEFAULT_LABELFONTSTYLE
			+ DEFAULT_LABELFONTSIZE;
	a->titlefont	= DEFAULT_TITLEFONTFAMILY
			+ DEFAULT_TITLEFONTSTYLE
			+ DEFAULT_TITLEFONTSIZE;

	a->thickness	= DEFAULT_THICKNESS;
	a->pointsize	= DEFAULT_POINTSIZE;

	a->xticks	= DEFAULT_XTICKS;
	a->yticks	= DEFAULT_YTICKS;
	a->zticks	= DEFAULT_ZTICKS;

	a->xmin		=  DBL_MAX;
	a->xmax		= -DBL_MAX;
	a->ymin		=  DBL_MAX;
	a->ymax		= -DBL_MAX;
	a->zmin		=  DBL_MAX;
	a->zmax		= -DBL_MAX;

	a->xviewmin	=  DBL_MAX;
	a->xviewmax	= -DBL_MAX;
	a->yviewmin	=  DBL_MAX;
	a->yviewmax	= -DBL_MAX;
	a->zviewmin	=  DBL_MAX;
	a->zviewmax	= -DBL_MAX;

	a->x_viewmin	= NULL;
	a->x_viewmax	= NULL;
	a->y_viewmin	= NULL;
	a->y_viewmax	= NULL;
	a->z_viewmin	= NULL;
	a->z_viewmax	= NULL;

	a->nplots	= 0;

	return a;
}

void free_plot3d_attribs_structure(Plot3Dattribs *a)
{

	if (a==NULL)
		return;

	testfree(a->id);

	free_range(a->xrange);
	free_range(a->yrange);

	free_expr(a->xgrid);
	free_expr(a->ygrid);
	
	free_expr(a->num_points);
	free_expr(a->num_contours);
	free_expr(a->tube_points);
	free_expr(a->tube_radius);

	free_expr(a->view_longitude);
	free_expr(a->view_colatitude);

	if (a->idx_overall!=0)
		glDeleteLists(a->idx_overall, 4);

	testfree(a->color);

	testfree(a->xlabel);
	testfree(a->ylabel);
	testfree(a->zlabel);

	free_expr(a->x_viewmin);
	free_expr(a->x_viewmax);
	free_expr(a->y_viewmin);
	free_expr(a->y_viewmax);
	free_expr(a->z_viewmin);
	free_expr(a->z_viewmax);

	testfree(a->title);

	free(a);
}

Plot3Dattribs *copy_plot3d_attribs_structure(Plot3Dattribs *a)
{
	Plot3Dattribs *t;

	if (a==NULL)
		return NULL;
	t = malloc(sizeof(Plot3Dattribs));

	t->id = copy_string(a->id);
	t->xrange = copy_range(a->xrange);
	t->yrange = copy_range(a->yrange);

	t->xgrid = copy_expr(a->xgrid);
	t->ygrid = copy_expr(a->ygrid);
	t->n = a->n;
	t->m = a->m;

	t->num_points = copy_expr(a->num_points);
	t->num_contours = copy_expr(a->num_contours);
	t->contours = a->contours;
	t->tube_points = copy_expr(a->tube_points);
	t->tube_radius = copy_expr(a->tube_radius);

	t->view_longitude = copy_expr(a->view_longitude);
	t->view_colatitude = copy_expr(a->view_colatitude);
	t->longitude = a->longitude;
	t->colatitude = a->colatitude;

	t->idx_overall = a->idx_overall;
	t->idx_objects = a->idx_objects;
	t->idx_bbox = a->idx_bbox;
	t->idx_title = a->idx_title;

	t->color = copy_string(a->color);
	t->rgbcolors[0] = a->rgbcolors[0];
	t->rgbcolors[1] = a->rgbcolors[1];
	t->rgbcolors[2] = a->rgbcolors[2];
	t->rgbcolors[3] = a->rgbcolors[3];

	t->xlabel = copy_string(a->xlabel);
	t->ylabel = copy_string(a->ylabel);
	t->zlabel = copy_string(a->zlabel);

	t->title = copy_string(a->title);
	t->titlefont = a->titlefont;
	t->titlewidth = a->titlewidth;
	
	t->axes = a->axes;
	t->gridstyle = a->gridstyle;
	t->scaling = a->scaling;
	t->shading = a->shading;
	t->style = a->style;
	t->coords = a->coords;
	t->thickness = a->thickness;
	t->pointsize = a->pointsize;

	t->font = a->font;
	t->axesfont = a->axesfont;
	t->labelfont = a->labelfont;

	t->xticks = a->xticks;
	t->yticks = a->yticks;
	t->zticks = a->zticks;

	t->xviewmin = a->xviewmin;
	t->xviewmax = a->xviewmax;
	t->yviewmin = a->yviewmin;
	t->yviewmax = a->yviewmax;
	t->zviewmin = a->zviewmin;
	t->zviewmax = a->zviewmax;

	t->x_viewmin = copy_expr(a->x_viewmin);
	t->x_viewmax = copy_expr(a->x_viewmax);
	t->y_viewmin = copy_expr(a->y_viewmin);
	t->y_viewmax = copy_expr(a->y_viewmax);
	t->z_viewmin = copy_expr(a->z_viewmin);
	t->z_viewmax = copy_expr(a->z_viewmax);

	t->xmin = a->xmin;
	t->xmax = a->xmax;
	t->ymin = a->ymin;
	t->ymax = a->ymax;
	t->zmin = a->zmin;
	t->zmax = a->zmax;

	t->cx = a->cx;
	t->cy = a->cy;
	t->cz = a->cz;

	t->xscale = a->xscale;
	t->yscale = a->yscale;
	t->zscale = a->zscale;

	t->radius = a->radius;

	t->nplots = a->nplots;

	return t;
}

/* creates a Plot3D structure; initializes fields with default values */ 
Plot3D *create_plot3d_structure(void)
{
	Plot3D *p;

	p = malloc(sizeof(Plot3D));

	p->textobj	= NULL;
	p->textalign	= ALIGN_CENTER;
	p->parg		= NULL;

	p->polygon_vertices = NULL;

	p->mesh		= NULL;
	p->normal	= NULL;
	p->curvepoints	= NULL;

	p->idx		= 0;

	p->attr		= NULL;
	
	p->xmin		=  DBL_MAX;
	p->xmax		= -DBL_MAX;
	p->ymin		=  DBL_MAX;
	p->ymax		= -DBL_MAX;
	p->zmin		=  DBL_MAX;
	p->zmax		= -DBL_MAX;

	p->fp		= NULL;
	p->bp		= NULL;
	p->next		= NULL;

	return p;
}

/* This routine probably never needs to be called from the outside.
   To make copies of plot structures use copy_plotlist() */
static Plot3D *copy_plot3d_structure(Plot3D *p)
{
	Plot3D *q;

	if (p==NULL)
		return NULL;

	q = malloc(sizeof(Plot3D));

	q->type = p->type;
	q->textobj = copy_textobject_list(p->textobj);
	q->textalign = p->textalign;
	q->polygon_vertices = copy_polygon_vertices(p->polygon_vertices);
	q->polygon_normal.x = p->polygon_normal.x;
	q->polygon_normal.y = p->polygon_normal.y;
	q->polygon_normal.z = p->polygon_normal.z;
	q->parg = copy_expr3(p->parg);
	q->mesh = copy_mesh(p->mesh, p->attr->m, p->attr->n);
	q->normal = copy_mesh(p->normal, p->attr->m, p->attr->n);
	q->curvepoints = copy_curvepoints(p->curvepoints, p->attr->n);
	q->idx = p->idx;

	q->xmin = p->xmin;
	q->xmax = p->xmax;
	q->ymin = p->ymin;
	q->ymax = p->ymax;
	q->zmin = p->zmin;
	q->zmax = p->zmax;

	q->attr = p->attr;

	q->fp = NULL;
	q->bp = NULL;
	q->next = NULL;

	return q;
}

/* frees the memory allocated to a linked list of Plotd3D structures */
void free_plotlist(Plot3D *p)
{
	Plot3D *pnext;

	while (p!=NULL) {
		pnext = p->fp;
		free_textobject_list(p->textobj);
		free_polygon_vertices(p->polygon_vertices);
		free_expr3(p->parg);
		free_mesh(p->mesh);
		free_mesh(p->normal);
		free_curvepoints(p->curvepoints);
		if (p->idx!=0)
			glDeleteLists(p->idx, 1);

		if (--p->attr->nplots==0)
			free_plot3d_attribs_structure(p->attr);

		free(p);

		p = pnext;
	}
}

Plot3D *copy_plotlist(Plot3D *p)
{
	Plot3D *q, *q0;
	int i;

	i = 0;
	q = q0 = NULL;
	while (p!=NULL) {
		if (q==NULL) {	/* first link */
			q = copy_plot3d_structure(p);
			q0 = q;	/* save starting address for returning */
		} else {
			q->fp = copy_plot3d_structure(p);
			q->fp->bp = q;
			q = q->fp;
		}

		if (i == 0)
			q->attr = copy_plot3d_attribs_structure(p->attr);

		if (++i == q->attr->nplots)
			i = 0;

		p = p->fp;
	}

	return q0;
}
