#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <setjmp.h>
#include <GL/gl.h>
#include "bsp.h"

void *gl2psMalloc(size_t nbytes);	/* defined in gl2ps.c */
float bspEpsilon;			/* value set in gl2ps.c */
extern jmp_buf gl2psJmpEnv;		/* value set in gl2ps.c */

static int simple_cmp(const void *A, const void *B)
{
	const bspObj *a = *(const bspObj **)A;
	const bspObj *b = *(const bspObj **)B;

	if (a->depth > b->depth)
		return -1;
	else if (a->depth < b->depth)
		return +1;
	else
		return 0;
}

#ifdef DEBUG
static void verify_sort(bspObj **arr, int n)
{
	int i;

	for (i=0; i<n-1; i++)
		if (arr[i]->depth < arr[i+1]->depth)
			printf("error: arr[%d] < arr[%d]\n", i, i+1);

	PRINTF(("verified sorting of array of %d objects\n", n));
}
#endif /* DEBUG */

void bspSimpleSort(bspNode *node)
{
	bspObj *obj;
	bspObj **a;
	size_t i, nobj;

	/* count number of objects */
	for (nobj=0, obj=node->obj; obj!=NULL; nobj++, obj=obj->next)
		;

	PRINTF(("simple-sorting %u objects\n", nobj));

	a = gl2psMalloc(nobj * sizeof *a);

	for (i=0, obj=node->obj; obj!=NULL; i++, obj=obj->next)
		a[i] = obj;

	qsort(a, nobj, sizeof *a, simple_cmp);

#ifdef DEBUG
	verify_sort(a, nobj);
#endif /* DEBUG */

	/* make the first element of the array a triangle object, if possible */
	if (a[0]->type != bspTriangleObj)
		for (i=1; i<nobj; i++)
			if (a[i]->type == bspTriangleObj) {
				bspObj *tmp = a[i];
				a[i] = a[0];
				a[0] = tmp;
				break;
			}

	for (i=0; i<nobj-1; i++)
		a[i]->next = a[i+1];
	a[i]->next = NULL;

	node->obj = a[0];

	free(a);
}

/* returns the dot product: a . b  */
double bspDotprod(const bspPoint a, const bspPoint b)
{
	double sum = 0.0;
	short i;

	for (i=0; i<3; i++)
		sum += a[i]*b[i];

	return sum;
}

/* computes the cross product:   c = a x b  */
static void crossprod(const bspPoint a, const bspPoint b, bspPoint c)
{
	c[0] = a[1] * b[2] - a[2] * b[1];
	c[1] = a[2] * b[0] - a[0] * b[2];
	c[2] = a[0] * b[1] - a[1] * b[0];
}

/* For a triangle with vertices V0, V1, V2 and computes its normal vector as:
   e1 = V1 - V0;
   e2 = V2 - V0;
   n = e1 x e2;
*/
static void compute_triangle_normal(const bspObj *obj, bspPlane plane)
{
	bspPoint e1, e2;
	short j;

	assert(obj->type == bspTriangleObj);

	for (j=0; j<3; j++) {
		e1[j] = obj->v[1][j] - obj->v[0][j];
		e2[j] = obj->v[2][j] - obj->v[0][j];
	}

	crossprod(e1, e2, (void*)plane);
}

/* The equation of the plane containing the triangle with vertices
   V0, V1, V2 and normal vector N is:  N . (X - V0) = 0.  Equivalently:
   N . X = N . V0.  This function computes the value of d = N . V0.
*/
static void compute_plane_offset(const bspObj *obj, bspPlane plane)
{
	plane[3] = bspDotprod(obj->v[0], (void*)plane);
}

static void get_object_plane(const bspObj *obj, bspPlane plane)
{
	switch(obj->type) {
		case bspTriangleObj:
			compute_triangle_normal(obj, plane);
			compute_plane_offset(obj, plane);
			break;
		default:
			fprintf(stderr, "--- in get_object_plane(): can handle triangle objects only\n");
			longjmp(gl2psJmpEnv, 1);
	}
}

static bspObj *line_split(const bspObj *obj, const bspPlane plane)
{
	bspObj *L1, *L2;
	bspPoint p;
	bspColor c;
	short i, l;
	double z[2], t;

	/* THIS CALCULATION REPEATS THAT IN line_vs_plane().
	   PERHAPS z[i] CAN BE SET AS A GLOBAL.
	*/
	for (i=0; i<2; i++)
		z[i] = bspDotprod(obj->v[i], (void*)plane) - plane[3];

	/* find the intersection point p1 and its color on the edge "jk" */
	t = z[0]/(z[0]-z[1]);

	for (l=0; l<3; l++)
		p[l] = obj->v[0][l] + (obj->v[1][l] - obj->v[0][l])*t;

	for (l=0; l<4; l++)
		c[l] = obj->rgba[0][l] + (obj->rgba[1][l] - obj->rgba[0][l])*t;

	/* sub-segment #1 */
	L1 = bspMakeObject(bspLineObj, 2);
	L1->linewidth = obj->linewidth;

	for (l=0; l<3; l++) {
		L1->v[0][l] = obj->v[0][l];
		L1->v[1][l] = p[l];
	}

	for (l=0; l<4; l++) {
		L1->rgba[0][l] = obj->rgba[0][l];
		L1->rgba[1][l] = c[l];
	}

#ifdef DEBUG
	L1->name = gl2psMalloc(32);
	sprintf(L1->name, "%s1", obj->name);
#endif /* DEBUG */

	/* sub-segment #2 */
	L2 = bspMakeObject(bspLineObj, 2);
	L2->linewidth = obj->linewidth;

	for (l=0; l<3; l++) {
		L2->v[0][l] = p[l];
		L2->v[1][l] = obj->v[1][l];
	}

	for (l=0; l<4; l++) {
		L2->rgba[0][l] = c[l];
		L2->rgba[1][l] = obj->rgba[1][l];
	}

#ifdef DEBUG
	L2->name = gl2psMalloc(32);
	sprintf(L2->name, "%s2", obj->name);
#endif /* DEBUG */

	L1->next = L2;

	PRINTF(("%s was split into %s %s\n", obj->name,
			L1->name, L1->next->name));
	return L1;
}

static bspCompare point_vs_plane(const bspObj *obj, const bspPlane plane)
{
	double z = bspDotprod(obj->v[0], (void*)plane) - plane[3];
	return z < 0 ? bspInBack : bspInFront;
}


/* returns one of bspCompare types depending on the position
   of a line segment with respect a partitioning plane
*/
static bspCompare line_vs_plane(const bspObj *obj, const bspPlane plane)
{
	double z;
	short c, d, i;

	d = 0;
	for (i=0; i<2; i++) {
		z = bspDotprod(obj->v[i], (void*)plane) - plane[3];
		c = z > bspEpsilon ? 1 : z < -bspEpsilon ? -1 : 0;
		d = 10*d + c+1;
	}

	switch (d) {		/* see `doc' for explanantion */
		case 11:
			return bspCoplanar;
		case 10:
		case 1:
		case 0:
			return bspInBack;
		case 12:
		case 21:
		case 22:
			return bspInFront;
		case 2:
		case 20:
			return bspSplit2;
		default:
			assert(0);	/* shouldn't be here */
	}
}

/* returns one of bspCompare types depending on the position
   of a triangle with respect a partitioning plane
*/
static bspCompare triangle_vs_plane(
		const bspObj *obj, 
		const bspPlane plane,
		short *vert)
{
	double z;
	short c, d, i;

	d = 0;
	for (i=0; i<3; i++) {
		z = bspDotprod(obj->v[i], (void*)plane) - plane[3];
		c = z > bspEpsilon ? 1 : z < -bspEpsilon ? -1 : 0;
		d = 10*d + c+1;
	}

	switch (d) {		/* see `doc' for explanantion */
		case 111:
			return bspCoplanar;
		case 0:
		case 100:
		case 10:
		case 1:
		case 11:
		case 101:
		case 110:
			return bspInBack;
		case 222:
		case 122:
		case 212:
		case 221:
		case 211:
		case 121:
		case 112:
			return bspInFront;
		case 120:
		case 102:
			*vert = 0;
			return bspSplit2;
		case 210:
		case 12:
			*vert = 1;
			return bspSplit2;
		case 201:
		case 21:
			*vert = 2;
			return bspSplit2;
		case 22:
		case 200:
			*vert = 0;
			return bspSplit3;
		case 202:
		case 20:
			*vert = 1;
			return bspSplit3;
		case 220:
		case 2:
			*vert = 2;
			return bspSplit3;
		default:
			assert(0);	/* shouldn't be here */
	}
}

static bspObj *triangle_split3(
		const bspObj *obj,
		const bspPlane plane,
		short vert)
{
	bspObj *T1, *T2, *T3;
	bspPoint p1, p2;
	bspColor c1, c2;
	short i, j, k, l;
	double z[3], t;

	/* THIS CALCULATION REPEATS THAT IN triangle_vs_plane().
	   PERHAPS z[i] CAN BE SET AS A GLOBAL.
	*/
	for (i=0; i<3; i++)
		z[i] = bspDotprod(obj->v[i], (void*)plane) - plane[3];

	i = vert;
	j = (i+1)%3;
	k = (i+2)%3;

	/* find the intersection point p1 and its color on the edge "ij" */
	t = z[i]/(z[i]-z[j]);

	for (l=0; l<3; l++)
		p1[l] = obj->v[i][l] + (obj->v[j][l] - obj->v[i][l])*t;

	for (l=0; l<4; l++)
		c1[l] = obj->rgba[i][l] + (obj->rgba[j][l] - obj->rgba[i][l])*t;

	/* find the intersection point p2 and its color on the edge "ik" */
	t = z[i]/(z[i]-z[k]);

	for (l=0; l<3; l++)
		p2[l] = obj->v[i][l] + (obj->v[k][l] - obj->v[i][l])*t;

	for (l=0; l<4; l++)
		c2[l] = obj->rgba[i][l] + (obj->rgba[k][l] - obj->rgba[i][l])*t;


	/* sub-triangle #1 */
	T1 = bspMakeObject(bspTriangleObj, 3);

	for (l=0; l<3; l++) {
		T1->v[i][l] = obj->v[i][l];
		T1->v[j][l] = p1[l];
		T1->v[k][l] = p2[l];
	}

	for (l=0; l<4; l++) {
		T1->rgba[i][l] = obj->rgba[i][l];
		T1->rgba[j][l] = c1[l];
		T1->rgba[k][l] = c2[l];
	}

#ifdef DEBUG
	T1->name = gl2psMalloc(32);
	sprintf(T1->name, "%s1", obj->name);
#endif /* DEBUG */

	/* sub-triangle #2 */
	T2 = bspMakeObject(bspTriangleObj, 3);

	for (l=0; l<3; l++) {
		T2->v[i][l] = p1[l];
		T2->v[j][l] = obj->v[j][l];
		T2->v[k][l] = p2[l];
	}

	for (l=0; l<4; l++) {
		T2->rgba[i][l] = c1[l];
		T2->rgba[j][l] = obj->rgba[j][l];
		T2->rgba[k][l] = c2[l];
	}

#ifdef DEBUG
	T2->name = gl2psMalloc(32);
	sprintf(T2->name, "%s2", obj->name);
#endif /* DEBUG */

	/* sub-triangle #3 */
	T3 = bspMakeObject(bspTriangleObj, 3);

	for (l=0; l<3; l++) {
		T3->v[i][l] = p2[l];
		T3->v[j][l] = obj->v[j][l];
		T3->v[k][l] = obj->v[k][l];
	}

	for (l=0; l<4; l++) {
		T3->rgba[i][l] = c2[l];
		T3->rgba[j][l] = obj->rgba[j][l];
		T3->rgba[k][l] = obj->rgba[k][l];
	}

#ifdef DEBUG
	T3->name = gl2psMalloc(32);
	sprintf(T3->name, "%s3", obj->name);
#endif /* DEBUG */

	T1->next = T2;
	T2->next = T3;

	PRINTF(("%s was split into %s %s %s\n", obj->name,
			T1->name, T1->next->name, T1->next->next->name));
	return T1;
}

static bspObj *triangle_split2(
		const bspObj *obj,
		const bspPlane plane,
		short vert)
{
	bspObj *T1, *T2;
	bspPoint p1;
	bspColor c1;
	short i, j, k, l;
	double z[3], t;

	/* THIS CALCULATION REPEATS THAT IN triangle_vs_plane().
	   PERHAPS z[i] CAN BE SET AS A GLOBAL.
	*/
	for (i=0; i<3; i++)
		z[i] = bspDotprod(obj->v[i], (void*)plane) - plane[3];

	i = vert;
	j = (i+1)%3;
	k = (i+2)%3;

	/* find the intersection point p1 and its color on the edge "jk" */
	t = z[j]/(z[j]-z[k]);

	for (l=0; l<3; l++)
		p1[l] = obj->v[j][l] + (obj->v[k][l] - obj->v[j][l])*t;

	for (l=0; l<4; l++)
		c1[l] = obj->rgba[j][l] + (obj->rgba[k][l] - obj->rgba[j][l])*t;

	/* sub-triangle #1 */
	T1 = bspMakeObject(bspTriangleObj, 3);

	for (l=0; l<3; l++) {
		T1->v[i][l] = obj->v[i][l];
		T1->v[j][l] = obj->v[j][l];
		T1->v[k][l] = p1[l];
	}

	for (l=0; l<4; l++) {
		T1->rgba[i][l] = obj->rgba[i][l];
		T1->rgba[j][l] = obj->rgba[j][l];
		T1->rgba[k][l] = c1[l];
	}

#ifdef DEBUG
	T1->name = gl2psMalloc(32);
	sprintf(T1->name, "%s1", obj->name);
#endif /* DEBUG */

	/* sub-triangle #2 */
	T2 = bspMakeObject(bspTriangleObj, 3);

	for (l=0; l<3; l++) {
		T2->v[i][l] = obj->v[i][l];
		T2->v[j][l] = p1[l];
		T2->v[k][l] = obj->v[k][l];
	}

	for (l=0; l<4; l++) {
		T2->rgba[i][l] = obj->rgba[i][l];
		T2->rgba[j][l] = c1[l];
		T2->rgba[k][l] = obj->rgba[k][l];
	}

#ifdef DEBUG
	T2->name = gl2psMalloc(32);
	sprintf(T2->name, "%s2", obj->name);
#endif /* DEBUG */

	T1->next = T2;

	PRINTF(("%s was split into %s %s\n", obj->name,
			T1->name, T1->next->name));
	return T1;
}

/* Add an object to an existing node's list of objects.
   If the list's first object is not a triangle but the incoming object
   is a triangle, stick the incoming object at the head of the list, else
   stick it at the end of the list.
*/
void bspObjAdd(bspNode *node, bspObj *obj)
{
	if (node->obj == NULL) {
		PRINTF(("starting an object list for a new node\n"));
		node->obj = obj;
		node->last = obj;
	} else if (node->obj->type != bspTriangleObj &&
			obj->type == bspTriangleObj) {
		PRINTF(("insering object at the head of the node\n"));
		obj->next = node->obj;
		node->obj = obj;
		if (node->last == NULL)
			node->last = obj;
	} else {
		PRINTF(("insering object at the tail of the node\n"));
		node->last->next = obj;
		node->last = obj;
	}
}

bspNode *bspMakeNode(void)
{
	bspNode *node = gl2psMalloc(sizeof *node);
	node->left = node->right = NULL;
	node->obj = node->last = NULL;
	return node;
}

static void push_left(bspObj *obj, bspNode *rootnode)
{
	if (rootnode->left==NULL)
		rootnode->left = bspMakeNode();

	bspObjAdd(rootnode->left, obj);
}

static void push_right(bspObj *obj, bspNode *rootnode)
{
	if (rootnode->right==NULL)
		rootnode->right = bspMakeNode();

	bspObjAdd(rootnode->right, obj);
}

static void free_text(Text *text)
{
	if (text == NULL)
		return;
	free(text->str);
	free(text->fontname);
	free(text);
}

bspObj *bspMakeObject(bspObjType type, short nverts)
{
	bspObj *obj = gl2psMalloc(sizeof *obj);
	obj->type = type;
	obj->nverts = nverts;
#ifdef DEBUG
	obj->name = NULL;
#endif /* DEBUG */
	obj->v = gl2psMalloc(nverts * sizeof(bspPoint));
	obj->rgba = gl2psMalloc(nverts * sizeof(bspColor));
	obj->linewidth = 0;
	obj->pointsize = 0;
	obj->text = NULL;
	obj->next = NULL;
	return obj;
}

static void free_object(bspObj *obj)
{
	PRINTF(("freeing object %s\n", obj->name));
#ifdef DEBUG
	free(obj->name);
#endif /* DEBUG */
	free(obj->v);
	free(obj->rgba);
	free_text(obj->text);
	free(obj);
}

static void free_node(bspNode *node)
{
	bspObj *obj;

	while (node->obj!=NULL) {
		obj = node->obj;
		node->obj = obj->next;
		free_object(obj);
	}
	free(node);
}

void bspFreeTree(bspNode *node)
{
	if (node==NULL)
		return;

	bspFreeTree(node->left);
	bspFreeTree(node->right);
	free_node(node);
}

/* Moves the point object in the node's left or right node and
   returns the original object.  Handles both bspPointObj and bspTextObj.
*/
static bspObj * process_point_object(bspObj *obj, bspNode *rootnode)
{
	PRINTF(("point or text `%s' ", obj->name));
	
	switch (point_vs_plane(obj, rootnode->plane)) {
		case bspInFront:
			PRINTF(("is in front of %s\n", rootnode->obj->name));
			push_right(obj, rootnode);
			break;
		case bspInBack:
			PRINTF(("is behind %s\n", rootnode->obj->name));
			push_left(obj, rootnode);
			break;
		default:
			/* not reached */
			assert(0);
	}

	return obj;
}

/* If the given line object is in front or behind the given node's
   partitioning plane, puts it in the node's left or right node and
   returns the object (the same of the received object.) If the line
   is intersected by the node's partitioning plane, returns a the of
   sub-segments.
*/
static bspObj *process_line_object(bspObj *obj, bspNode *rootnode)
{
	bspObj *new = obj;

	PRINTF(("line `%s' ", obj->name));

	switch (line_vs_plane(obj, rootnode->plane)) {

		case bspCoplanar:
			PRINTF(("is in the plane of %s\n", rootnode->obj->name));
			PRINTF(("ASSUMING "));
			/* FALL THROUGH */
		case bspInFront:
			PRINTF(("is in front of %s\n", rootnode->obj->name));
			push_right(obj, rootnode);
			break;
		case bspInBack:
			PRINTF(("is behind %s\n", rootnode->obj->name));
			push_left(obj, rootnode);
			break;
		case bspSplit2:
			PRINTF(("is split by %s\n", rootnode->obj->name));
			new = line_split(obj, rootnode->plane);
			break;
		case bspSplit3:
			assert(0);	/* shouldn't be here */
	}

	return new;
}

/* If the given triangle object is in front or behind the given node's
   partitioning plane, puts it in the node's left or right node and
   returns the object (the same of the received object.) If the triangle
   is intersected by the node's partitioning plane, returns the list of
   sub-triangles.
*/
static bspObj *process_triangle_object(bspObj *obj, bspNode *rootnode)
{
	bspObj *new = obj;
	short vert;

	PRINTF(("triangle `%s' ", obj->name));

	switch (triangle_vs_plane(obj, rootnode->plane, &vert)) {

		case bspCoplanar:
			PRINTF(("is in the plane of %s\n", rootnode->obj->name));
			PRINTF(("ASSUMING "));
			/* FALL THROUGH */
		case bspInFront:
			PRINTF(("is in front of %s\n", rootnode->obj->name));
			push_right(obj, rootnode);
			break;
		case bspInBack:
			PRINTF(("is behind %s\n", rootnode->obj->name));
			push_left(obj, rootnode);
			break;
		case bspSplit2:
			PRINTF(("is split in two at vert=%d by %s\n",
						vert, rootnode->obj->name));
			new = triangle_split2(obj, rootnode->plane, vert);
			break;
		case bspSplit3:
			PRINTF(("is split in three (vert=%d) by %s\n",
						vert, rootnode->obj->name));
			new = triangle_split3(obj, rootnode->plane, vert);
			break;
	}

	return new;
}

static void process_object_list(bspObj *obj, bspNode *rootnode)
{
	bspObj *new=NULL, *curr;

	while (obj!=NULL) {

		/* remove object from the linked list */
		curr = obj;
		obj = obj->next;
		curr->next = NULL;

		switch (curr->type) {
			case bspTextObj:
			case bspPointObj:
				new = process_point_object(curr, rootnode);
				break;
			case bspLineObj:
				new = process_line_object(curr, rootnode);
				break;
			case bspTriangleObj:
				new = process_triangle_object(curr, rootnode);
				break;
		}

		if (new!=curr)		/* current object was split */
			process_object_list(new, rootnode);
	}
}

void bspProduceTree(bspNode *rootnode)
{
	assert(rootnode!=NULL && rootnode->obj!=NULL && rootnode->obj->type == bspTriangleObj);

	get_object_plane(rootnode->obj, rootnode->plane);

	PRINTF(("ROOT IS `%s' WITH PLANE (%g, %g, %g) %g\n",
				rootnode->obj->name,
				rootnode->plane[0],
				rootnode->plane[1],
				rootnode->plane[2],
				rootnode->plane[3]));

	process_object_list(rootnode->obj->next, rootnode);
	rootnode->obj->next = NULL;

	if (rootnode->left != NULL
			&& rootnode->left->obj->type == bspTriangleObj) {

		PRINTF(("calling bspProduceTree() with left of %s which is %s\n",
				rootnode->obj->name, rootnode->left->obj->name));
		bspProduceTree(rootnode->left);
	}

	if (rootnode->right != NULL
			&& rootnode->right->obj->type == bspTriangleObj) {

		PRINTF(("calling bspProduceTree() with right of %s which is %s\n",
				rootnode->obj->name, rootnode->right->obj->name));
		bspProduceTree(rootnode->right);
	}

}
