#include <stdio.h>
#include <assert.h>
#include <stdlib.h>	/* for free() */
#include <math.h>	/* for floor() */
#include <GL/gl.h>
#include "gl2ps.h"
#include "dynagraph.h"

static GLfloat	mat_shininess = 128.0;
static GLfloat	mat_specular[] = { 1.0, 1.0, 1.0, ALPHA };
extern GLfloat	blackColor[];
extern GLfloat	*bgColor, *fgColor;
extern char	*prognam;
extern gl2psContext *gl2ps;
extern double zoom_scale;

/* Determined experimentally to minimize "stitching".  The Red Book recommends
   smaller values.
*/
#define POLYGON_OFFSET_FACTOR	1.0
#define POLYGON_OFFSET_UNITS	10.0

#ifdef MAPLE_STYLE_ZHUE
/* The color scheme here is modeled after Maple's ZHUE rendering colors.
   We do not use this by default because Matlab's ZHUE rendering colors,
   described below, produces better pictures.
*/

#define REDHUE(s)	(huefunc((s)+1.0/3))
#define GREENHUE(s)	(huefunc((s)+2.0/3))
#define BLUEHUE(s)	(huefunc((s)))

/* A piecewise-linear periodic function R -> [0,1] with period 1.
   Its graph on [0,1] connects the points (0,0), (1/6, 1), (1/2, 1),
   (2/3, 0), (1,0), and extens periodically to R.  This function is
   used for drawing objects with the HUE coloring scheme.
*/
static float huefunc(float s)
{
	s = s - floor(s);

	if (s < 1.0/6)
		return 6*s;
	else if (s < 3.0/6)
		return 1.0;
	else if (s < 4.0/6)
		return 4 - 6*s;
	else
		return 0;
}

#else	/* Matlab's ZHUE */

#define REDHUE(s)       (huefunc((s)-23.0/64))
#define GREENHUE(s)     (huefunc((s) -7.0/64))
#define BLUEHUE(s)      (huefunc((s) +9.0/64))

/* This defines a continuous, piecewise-linear function defined on
   the entire real line.  Its graph is trapezoid-shaped and connects
   the points (0,0) (16/64,1), (33/64,1), (49/64,0).  The function
   is zero outside the support of the trapezoid.
   In combination with REDUE, GREENHUE, BLUEHUE macros defined above,
   produces an RGB scale used for our ZHUE rendering which coincides
   with Matlab's JET colormap.
*/
static float huefunc(float s)
{
	if (s < 0)
		return 0.0;
	else if (s < 16.0/64)
		return 4.0*s;
	else if (s < 33.0/64)
		return 1.0;
	else if (s < 49.0/64)
		return 1.0 - 4.0*(s-33.0/64);
	else
		return 0.0;
}

#endif /* MAPLE_STYLE_ZHUE */

static GLfloat *get_color(Plot3D *p, double x, double y, double z)
{
	static GLfloat v[] = { 0, 0, 0, ALPHA };	/* v[3] will remain fixed */
	double xmin = p->attr->xmin;
	double ymin = p->attr->ymin;
	double zmin = p->attr->zmin;
	double xmax = p->attr->xmax;
	double ymax = p->attr->ymax;
	double zmax = p->attr->zmax;
	float a=0.15, b=0.9;
	float c=0.4, d=0.2;
	float R, G, B;
	unsigned int shading;

	shading = p->attr->shading&~locked;

	/* scale the values of x, y, z to the range 0..1 */
	x = (xmax-xmin<1e-5) ? 0.5 : (x-xmin)/(xmax-xmin);
	y = (ymax-ymin<1e-5) ? 0.5 : (y-ymin)/(ymax-ymin);
	z = (zmax-zmin<1e-5) ? 0.5 : (z-zmin)/(zmax-zmin);

	switch (shading) {
		case XYZSHADING:
			/* set color values in the range [a, b] */
			R = a + (b-a)*x;
			G = a + (b-a)*y;
			B = a + (b-a)*z;
			break;
		case XYSHADING:
			R = 1.0;
			G = x;
			B = y;
			break;
		case ZSHADING:
			R = c + (1.0-c)*z;	/* R ranges in [c, 1] */
			G = 0.0;
			B = 1.0 - (1.0-d)*z;	/* B ranges in [d, 1] */
			break;
		case ZHUE:
			R = REDHUE(z);
			G = GREENHUE(z);
			B = BLUEHUE(z);
			break;
		case ZGRAYSCALE:	/* numbers mimicing Maple's */
			R = (0x49 + (0xF7-0x49)*z)/0xFF;
			G = (0x4D + (0xF3-0x4D)*z)/0xFF;
			B = R;
			break;
		case NOSHADING:
			R = 1.0;
			G = 1.0;
			B = 1.0;
			break;
		default:
			/* anything else gets colored as in XYZ shading */
			R = a + (b-a)*x;
			G = a + (b-a)*y;
			B = a + (b-a)*z;
			break;
	}

	v[0] = R; v[1] = G; v[2] = B;
	return v;
}

void set_vertex_color(Plot3D *p, double x, double y, double z)
{
	glColor4fv(p->attr->color==NULL ? get_color(p, x, y, z) :
			p->attr->rgbcolors);	
}

/* Draws surface as a mesh of points */
static void construct_point_mesh(Plot3D *p)
{
	int i, j;
	GLdouble x, y, z;

	glBegin(GL_POINTS);
	for (i=0; i<p->attr->m; i++)
		for (j=0; j<p->attr->n; j++) {
			if (IS_SINGULAR(p->mesh[i][j]))
				continue;
			x = p->mesh[i][j].x;
			y = p->mesh[i][j].y;
			z = p->mesh[i][j].z;
			set_vertex_color(p, x, y, z);
			glVertex3f(x,y,z);
		}
	glEnd();
}

/* Draws the outline of a rectangular mesh.
   Objects are drawn in the fixed color specified in the "color" argument
   (used for drawing an overlaid grid.)  If that argument is NULL, then
   the color of each vertex is determined by calling the set_vertex_color()
   function.
*/
static void construct_rectangular_mesh(Plot3D *p, GLfloat *color, unsigned int gridstyle)
{
	int i, j;
	GLdouble x, y, z;

	if (color!=NULL)
		glColor4fv(color);

	if (gridstyle==XY_GRIDLINES || gridstyle==Y_GRIDLINES) {
		for (i=0; i<p->attr->m; i++) {
			glBegin(GL_LINE_STRIP);
			for (j=0; j<p->attr->n; j++) {
				if (IS_SINGULAR(p->mesh[i][j])) {
					glEnd();
					glBegin(GL_LINE_STRIP);
					continue;
				}
				x = p->mesh[i][j].x;
				y = p->mesh[i][j].y;
				z = p->mesh[i][j].z;
				if (color==NULL)
					set_vertex_color(p, x, y, z);
				glVertex3f(x,y,z);
			}
			glEnd();
		}
	}

	if (gridstyle==XY_GRIDLINES || gridstyle==X_GRIDLINES) {
		for (j=0; j<p->attr->n; j++) {
			glBegin(GL_LINE_STRIP);
			for (i=0; i<p->attr->m; i++) {
				if (IS_SINGULAR(p->mesh[i][j])) {
					glEnd();
					glBegin(GL_LINE_STRIP);
					continue;
				}
				x = p->mesh[i][j].x;
				y = p->mesh[i][j].y;
				z = p->mesh[i][j].z;
				if (color==NULL)
					set_vertex_color(p, x, y, z);
				glVertex3f(x,y,z);
			}
			glEnd();
		}
	}
}

static void construct_polygon_outline(Plot3D *p, GLfloat *color)
{
	PolygonVertex *vertex;

	if (color!=NULL)
		glColor4fv(color);

	glBegin(GL_LINE_LOOP);
	for (vertex=p->polygon_vertices; vertex!=NULL; vertex=vertex->fp) {
		double x = vertex->numeric.x;
		double y = vertex->numeric.y;
		double z = vertex->numeric.z;
		glVertex3f(x,y,z);
	}
	glEnd();
}

static void construct_polygon(Plot3D *p, GLfloat *color)
{
	PolygonVertex *vertex;

	glBegin(GL_TRIANGLE_FAN);
	glNormal3f(p->polygon_normal.x,
			p->polygon_normal.y, p->polygon_normal.z); 

	for (vertex=p->polygon_vertices; vertex!=NULL; vertex=vertex->fp) {
		double x = vertex->numeric.x;
		double y = vertex->numeric.y;
		double z = vertex->numeric.z;
		if (color==NULL)
			set_vertex_color(p, x, y, z);
		glVertex3f(x,y,z);
	}
	glEnd();
}

/* With glPolygonMode(GL_FILL) draws surface with filled triangles, no lines.
   With glPolygonMode(GL_LINE) draws surface as a triangular mesh outline,
   not filled.
   Objects are drawn in the fixed color specified in the "color" argument
   (used for drawing an overlaid grid.)  If that argument is NULL, then
   the color of each vertex is determined by calling the set_vertex_color()
   function.
*/
static void construct_surface(Plot3D *p, GLfloat *color)
{
	double x, y, z;
	int i, j;
	int jumping = 0;

	if (color!=NULL)
		glColor4fv(color);

	if (p->type == POLYGON_PLOT) {
		construct_polygon(p, color);
		return;
	}

	for (i=0; i<p->attr->m-1; i++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (j=0; j<p->attr->n; j++) {
			jumping = 0;
			if (IS_SINGULAR(p->mesh[i][j])) {
				glEnd();
				glBegin(GL_TRIANGLE_STRIP);
				jumping = 1;
				goto next;
			}
			x = p->mesh[i][j].x;
			y = p->mesh[i][j].y;
			z = p->mesh[i][j].z;
			if (color==NULL)
				set_vertex_color(p, x, y, z);
			glNormal3f(	p->normal[i][j].x,
					p->normal[i][j].y,
					p->normal[i][j].z);
			glVertex3f(x,y,z);

next:
			if (IS_SINGULAR(p->mesh[i+1][j])) {
				glEnd();
				glBegin(GL_TRIANGLE_STRIP);
				continue;
			}
			x = p->mesh[i+1][j].x;
			y = p->mesh[i+1][j].y;
			z = p->mesh[i+1][j].z;
			if (color==NULL)
				set_vertex_color(p, x, y, z);
			glNormal3f(	p->normal[i+1][j].x,
					p->normal[i+1][j].y,
					p->normal[i+1][j].z);
			/* need this to maintain correct orientation */
			if (jumping)
				glVertex3f(x,y,z);
			glVertex3f(x,y,z);
		}
		glEnd();
	}
}

static void construct_curve(Plot3D *curve)
{
	unsigned int style;
	double x, y, z;
	int i;

	style = curve->attr->style&~locked;

	style == POINT ?  glBegin(GL_POINTS) : glBegin(GL_LINE_STRIP);

	for (i=0; i<curve->attr->n; i++) {
		if (IS_SINGULAR(curve->curvepoints[i])) {
			glEnd();
			style == POINT ?  glBegin(GL_POINTS) : glBegin(GL_LINE_STRIP);
			continue;
		}
		x = curve->curvepoints[i].x;
		y = curve->curvepoints[i].y;
		z = curve->curvepoints[i].z;

		set_vertex_color(curve, x, y, z);
		glVertex3f(x,y,z);
	}

	glEnd();
}

static int decode_size(unsigned int w)
{
	int i=0;

	while (w) {
		w >>= 1;
		i++;
	}
	return i-1;
}

/* TEXT OBJECT */
/* Note 1: I don't know how to adjust text objects according to the ALIGN
   options.  In the current setup the ALIGN options are ignored.

   Note 2: I don't think that text objects should be affected by lighting.
   Theforefore I am drawing here with no lighting.  (It would be just as easy
   to draw them with lighting.)
*/
static void compile_textobject(Plot3D *p)
{
	TextObj *t = p->textobj;
	Fontdesc *f = generate_font(p->attr->font&~locked);
	GLdouble mvmatrix[16], pjmatrix[16];
	GLint viewport[4];
	GLfloat *v;
	char *psfontname;
	int psfontsize;

	assert(f!=NULL);

	if (gl2ps) get_PostscriptFont(p->attr->font&~locked,
				&psfontname, &psfontsize);

	if (!gl2ps) glNewList(p->idx, GL_COMPILE);
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		v = p->attr->color==NULL ? fgColor : p->attr->rgbcolors;
		glColor4fv(v);

		glGetIntegerv(GL_VIEWPORT, viewport);
		glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
		glGetDoublev(GL_PROJECTION_MATRIX, pjmatrix);

		while (t!=NULL) {
			glRasterPos3f(t->x, t->y, t->z);
			if (gl2ps) gl2psText(t->text, psfontname, psfontsize);
			print_string(t->text, f->glbase);
			t=t->fp;
		}

		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* SPACEVURVE
   Note:  I don't think that a curve object should be affected by lighting.
   Theforefore I am drawing here with no lighting.  (It would be just as easy
   to draw it with lighting.)
*/
static void compile_spacecurve(Plot3D *curve)
{
	if (!gl2ps) glNewList(curve->idx, GL_COMPILE);

		/* turn antialiasing on, based on the menu setting */
		if (antialias()) {
			glEnable(GL_POINT_SMOOTH);
			glEnable(GL_LINE_SMOOTH);
			glEnable(GL_BLEND);
		}

		glPointSize(decode_size(curve->attr->pointsize&~locked));
		gl2psPointSize(decode_size(curve->attr->pointsize&~locked));

		glLineWidth(decode_size(curve->attr->thickness&~locked));
		gl2psLineWidth(decode_size(curve->attr->thickness&~locked));

		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		construct_curve(curve);
		glPopAttrib();	/* GL_LIGHTING_BIT */
		glDisable(GL_POINT_SMOOTH);
		glDisable(GL_LINE_SMOOTH);
		glDisable(GL_BLEND);

	if (!gl2ps) glEndList();
}

/* style = PATCHNOGRID */
static void compile_patch_surface(Plot3D *p)
{
	if (!gl2ps) glNewList(p->idx, GL_COMPILE);
		glPushAttrib(GL_LIGHTING_BIT);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		construct_surface(p, NULL);
		glDisable(GL_COLOR_MATERIAL);
		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* style = PATCH */
static void compile_patch_and_grid_surface(Plot3D *p)
{
	unsigned int gridstyle = p->attr->gridstyle&~locked;

	if (!gl2ps) glNewList(p->idx, GL_COMPILE);

		glLineWidth(decode_size(p->attr->thickness&~locked));
		gl2psLineWidth(decode_size(p->attr->thickness&~locked));

		/* draw the mesh part of the surface first */
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		/* Here we choose to draw the grid lines in black.
		   Alternatively we could've drawn in fgColor. */
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		if (p->type == POLYGON_PLOT)
			construct_polygon_outline(p, blackColor);
		else {
			if (gridstyle == TRIANGULAR_GRIDLINES)
				construct_surface(p, blackColor);
			else
				construct_rectangular_mesh(p, blackColor, gridstyle);
		}
		glPopAttrib();	/* GL_LIGHTING_BIT */

		/* draw the filled part of the surface next */
		glPushAttrib(GL_LIGHTING_BIT);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(POLYGON_OFFSET_FACTOR, POLYGON_OFFSET_UNITS);
		construct_surface(p, NULL);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glDisable(GL_COLOR_MATERIAL);
		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* style = HIDDEN */
static void compile_hiddenline_surface(Plot3D *p)
{
	unsigned int gridstyle;

	gridstyle = p->attr->gridstyle&~locked;

	if (!gl2ps) glNewList(p->idx, GL_COMPILE);

		glLineWidth(decode_size(p->attr->thickness&~locked));
		gl2psLineWidth(decode_size(p->attr->thickness&~locked));

		/* draw the full wireframe representation of the object first */
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		if (gridstyle == TRIANGULAR_GRIDLINES)
			construct_surface(p, NULL);
		else
			construct_rectangular_mesh(p, NULL, gridstyle);

		/* now fill the space between the gridlines with bg color */
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(POLYGON_OFFSET_FACTOR, POLYGON_OFFSET_UNITS);
		construct_surface(p, bgColor);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* style = CONTOUR */
static void compile_contour_surface(Plot3D *p)
{
	if (!gl2ps) glNewList(p->idx, GL_COMPILE);
		glLineWidth(decode_size(p->attr->thickness&~locked));
		gl2psLineWidth(decode_size(p->attr->thickness&~locked));
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		construct_contours(p, NULL);
		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* style = PATCHCONTOUR */
static void compile_patch_and_contour_surface(Plot3D *p)
{
	if (!gl2ps) glNewList(p->idx, GL_COMPILE);

		glLineWidth(decode_size(p->attr->thickness&~locked));
		gl2psLineWidth(decode_size(p->attr->thickness&~locked));

		/* draw the contour lines first */
		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		/* Here we choose to draw the contour lines in black.
		   Alternatively we could've drawn in fgColor. */
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		construct_contours(p, blackColor);
		glPopAttrib();	/* GL_LIGHTING_BIT */

		/* draw the filled part of the surface next */
		glPushAttrib(GL_LIGHTING_BIT);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(POLYGON_OFFSET_FACTOR, POLYGON_OFFSET_UNITS);
		construct_surface(p, NULL);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glDisable(GL_COLOR_MATERIAL);
		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}

/* style = WIREFRAME */
static void compile_wireframe_surface(Plot3D *p)
{
	unsigned int gridstyle = p->attr->gridstyle&~locked;

	if (!gl2ps) glNewList(p->idx, GL_COMPILE);

		glLineWidth(decode_size(p->attr->thickness&~locked));
		gl2psLineWidth(decode_size(p->attr->thickness&~locked));

		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		if (gridstyle == TRIANGULAR_GRIDLINES)
			construct_surface(p, NULL);
		else
			construct_rectangular_mesh(p, NULL, gridstyle);
		glPopAttrib();	/* GL_LIGHTING_BIT */

	if (!gl2ps) glEndList();
}

/* style = POINT */
static void compile_point_surface(Plot3D *p)
{
	if (!gl2ps) glNewList(p->idx, GL_COMPILE);

		glPointSize(decode_size(p->attr->pointsize&~locked));
		gl2psPointSize(decode_size(p->attr->pointsize&~locked));

		glPushAttrib(GL_LIGHTING_BIT);
		glDisable(GL_LIGHTING);
		
		if (antialias()) {
			glEnable(GL_BLEND);
			glEnable(GL_LINE_SMOOTH);
			glEnable(GL_POINT_SMOOTH);
		}

		construct_point_mesh(p);

		glDisable(GL_BLEND);
		glDisable(GL_LINE_SMOOTH);
		glDisable(GL_POINT_SMOOTH);

		glPopAttrib();	/* GL_LIGHTING_BIT */
	if (!gl2ps) glEndList();
}


/*
We call XParseColor() to translate a color name to the corresponding RGB
values (which lie in the range 0 to 65535 inclusive).
We could use fl_lookup_RGBcolor() instead, but we do it here our own way.
Our function returns -1 on failure and 0 on success.

Note:
In XFree86's Xcolor(3x) man page we read:

       The XColor structure contains:

       typedef struct {
            unsigned long pixel;
            unsigned short red, green, blue;
            char flags;
            char pad;
       } XColor;

       The red, green, and blue values are always in the range 0
       to 65535 inclusive, independent of the number of bits
       actually used in the display hardware.  The server scales
       these values down to the range used by the hardware.
       Black is represented by (0,0,0), and white is represented
       by (65535,65535,65535).


*/
static int lookup_RGBcolor(const char *colname,
	unsigned short *r, unsigned short *g, unsigned short *b)
{
	XColor xc;

	if (XParseColor(fl_display, fl_state[fl_vmode].colormap,
				colname,  &xc) == 0)
		return -1;

	*r = xc.red;
	*g = xc.green;
	*b = xc.blue;

	return 0;
}

static void compile_object(Plot3D *p)
{
	unsigned int style;

	/* if option color=name is present, process it now */
	if (p->attr->color!=NULL && p->attr->rgbcolors[0]==-1) {
		unsigned short r, g, b;
		if (lookup_RGBcolor(p->attr->color, &r, &g, &b) < 0 ) {
			printf("\twarning: unrecognized color name %s\n",
					p->attr->color);
			free(p->attr->color);
			p->attr->color = NULL;
		} else {
			p->attr->rgbcolors[0] = (double)r / 65535;
			p->attr->rgbcolors[1] = (double)g / 65535;
			p->attr->rgbcolors[2] = (double)b / 65535;
		}
	}

	if (p->type == CURVE_PLOT) {	/* it's a spacecurve! */
		compile_spacecurve(p);
		return;
	}

	if (p->type == TEXT_PLOT) {
		compile_textobject(p);
		return;
	}

	style = p->attr->style&~locked;

	switch(style) {
		case POINT:
			compile_point_surface(p);
			break;
		case HIDDEN:
			compile_hiddenline_surface(p);
			break;
		case PATCH:
			compile_patch_and_grid_surface(p);
			break;
		case WIREFRAME:
			compile_wireframe_surface(p);
			break;
		case PATCHNOGRID:
			compile_patch_surface(p);
			break;
		case CONTOUR:
			compile_contour_surface(p);
			break;
		case PATCHCONTOUR:
			compile_patch_and_contour_surface(p);
			break;
		default:
			compile_patch_surface(p);	/* not reached */
			break;
	}
}

static void compile_objects(Plot3D *p)
{
	GLdouble clip_x_lo[4], clip_x_hi[4];
	GLdouble clip_y_lo[4], clip_y_hi[4];
	GLdouble clip_z_lo[4], clip_z_hi[4];
	float eps = 1e-4;
	Plot3D *q;
	GLuint idx;

	/* Set up clip planes. 
	   Clipping the object is needed only when view=... is specified
	   but we do it in all cases, anyway.
	   The eps adjustment is needed to make sure that we do not clip
	   a face of the object which happens to lie exactly in the clip
	   plane.  Actually, the adjustment should be a function of the 
	   size of the object, so we are being quite simple-minded here.

	   Each of the six clipping planes specified below is the representation
	   of a plane of the form Ax + By + Cz + D = 0.
	*/
	clip_x_lo[0] = +1;
	clip_x_lo[1] = 0;
	clip_x_lo[2] = 0;
	clip_x_lo[3] = -p->attr->xmin+eps;

	clip_x_hi[0] = -1;
	clip_x_hi[1] = 0;
	clip_x_hi[2] = 0;
	clip_x_hi[3] = +p->attr->xmax+eps;

	clip_y_lo[0] = 0;
	clip_y_lo[1] = +1;
	clip_y_lo[2] = 0;
	clip_y_lo[3] = -p->attr->ymin+eps;

	clip_y_hi[0] = 0;
	clip_y_hi[1] = -1;
	clip_y_hi[2] = 0;
	clip_y_hi[3] = +p->attr->ymax+eps;

	clip_z_lo[0] = 0;
	clip_z_lo[1] = 0;
	clip_z_lo[2] = +1;
	clip_z_lo[3] = -p->attr->zmin+eps;

	clip_z_hi[0] = 0;
	clip_z_hi[1] = 0;
	clip_z_hi[2] = -1;
	clip_z_hi[3] = +p->attr->zmax+eps;

	if (!gl2ps) {
		/* Allocate display lists for this plot:
		   one per graph
		   one for the merged set of graphs
		   one for the title
		   one for the bounding box
		   one for the overall scene whcih includes everything above
		*/
		idx = glGenLists(4 + p->attr->nplots);
		if (idx==0) {
			printf("%s: glGenLists() failed! %s: line %d\n",
					prognam, __FILE__, __LINE__);
			exit(EXIT_FAILURE);
		}

		p->attr->idx_overall	= idx++;
		p->attr->idx_objects	= idx++;
		p->attr->idx_bbox	= idx++;
		p->attr->idx_title	= idx++;

		for (q=p; q!=NULL; q= q->fp) {
			q->idx = idx++;
			compile_object(q);
		}
	}
	
	if (!gl2ps) glNewList(p->attr->idx_objects, GL_COMPILE);

	glPushMatrix();

	glClipPlane(GL_CLIP_PLANE0, clip_x_lo);
	glClipPlane(GL_CLIP_PLANE1, clip_x_hi);
	glClipPlane(GL_CLIP_PLANE2, clip_y_lo);
	glClipPlane(GL_CLIP_PLANE3, clip_y_hi);
	glClipPlane(GL_CLIP_PLANE4, clip_z_lo);
	glClipPlane(GL_CLIP_PLANE5, clip_z_hi);

	glEnable(GL_CLIP_PLANE0);
	glEnable(GL_CLIP_PLANE1);
	glEnable(GL_CLIP_PLANE2);
	glEnable(GL_CLIP_PLANE3);
	glEnable(GL_CLIP_PLANE4);
	glEnable(GL_CLIP_PLANE5);

	if (!gl2ps) {
		/* draw point/curve objects after others, so that the antialiasing
		   works correctly */
		for (q=p; q!=NULL; q= q->fp)
			if (!(q->type==CURVE_PLOT || (q->attr->style&~locked)==POINT))
				glCallList(q->idx);

		for (q=p; q!=NULL; q= q->fp)
			if (q->type==CURVE_PLOT || (q->attr->style&~locked)==POINT)
				glCallList(q->idx);
	} else
		for (q=p; q!=NULL; q= q->fp)
			compile_object(q);


	glDisable(GL_CLIP_PLANE0);
	glDisable(GL_CLIP_PLANE1);
	glDisable(GL_CLIP_PLANE2);
	glDisable(GL_CLIP_PLANE3);
	glDisable(GL_CLIP_PLANE4);
	glDisable(GL_CLIP_PLANE5);

	glPopMatrix();
	if (!gl2ps) glEndList();
}

#ifdef BROKEN_GL_NORMALIZE

static void restore_polygon_normals(Plot3D *p)
{
	double nx, ny, nz, len;

	nx = p->polygon_normal.x;
	ny = p->polygon_normal.y;
	nz = p->polygon_normal.z;
	len =  sqrt(nx*nx + ny*ny + nz*nz) / zoom_scale;
	p->polygon_normal.x /= len;
	p->polygon_normal.y /= len;
	p->polygon_normal.z /= len;
}

static void restore_object_normals(Plot3D *p)
{
	double nx, ny, nz, len;
	int i, j;

	for (i=0; i<p->attr->m; i++)
		for (j=0; j<p->attr->n; j++) {
			nx = p->normal[i][j].x;
			ny = p->normal[i][j].y;
			nz = p->normal[i][j].z;
			len =  sqrt(nx*nx + ny*ny + nz*nz) / zoom_scale;
			p->normal[i][j].x /= len;
			p->normal[i][j].y /= len;
			p->normal[i][j].z /= len;
		}
}

static void restore_normals(Plot3D *p)
{
	Plot3D *q;

	for (q=p; q!=NULL; q=q->fp)
		if (q->normal!=NULL)
			restore_object_normals(q);
		else if (q->type == POLYGON_PLOT)
			restore_polygon_normals(q);
}

static void scale_polygon_normals(Plot3D *p, double a, double b, double c)
{
	double nx, ny, nz, len;

	a *= zoom_scale;
	b *= zoom_scale;
	c *= zoom_scale;

	nx = p->polygon_normal.x / a;
	ny = p->polygon_normal.y / b;
	nz = p->polygon_normal.z / c;
	len =  sqrt(nx*nx + ny*ny + nz*nz);

	p->polygon_normal.x /= len;
	p->polygon_normal.y /= len;
	p->polygon_normal.z /= len;
}

static void scale_object_normals(Plot3D *p, double a, double b, double c)
{
	double nx, ny, nz, len;
	int i, j;

	a *= zoom_scale;
	b *= zoom_scale;
	c *= zoom_scale;

	for (i=0; i<p->attr->m; i++)
		for (j=0; j<p->attr->n; j++) {
			nx = p->normal[i][j].x / a;
			ny = p->normal[i][j].y / b;
			nz = p->normal[i][j].z / c;
			len =  sqrt(nx*nx + ny*ny + nz*nz);
			p->normal[i][j].x /= len;
			p->normal[i][j].y /= len;
			p->normal[i][j].z /= len;
		}
}

void scale_normals(Plot3D *p)
{
	double a = p->attr->xscale;
	double b = p->attr->yscale;
	double c = p->attr->zscale;
	Plot3D *q;

	for (q=p; q!=NULL; q=q->fp)
		if (q->normal!=NULL)
			scale_object_normals(q, a, b, c);
		else if (q->type == POLYGON_PLOT)
			scale_polygon_normals(q, a, b, c);
}
#endif /* BROKEN_GL_NORMALIZE */

void compile_scene(Plot3D *p)
{
	if (!gl2ps) {
#ifdef BROKEN_GL_NORMALIZE
		if ((p->attr->scaling&~locked)==UNCONSTRAINED)
			scale_normals(p);
		else if ((p->attr->scaling&~locked)==CONSTRAINED)
			restore_normals(p);
#endif /* BROKEN_GL_NORMALIZE */

		compile_objects(p);

		compile_boundingbox(p->attr);

		if (p->attr->title!=NULL)
			compile_title(p->attr);

		glNewList(p->attr->idx_overall, GL_COMPILE);
		glPushMatrix();

		if ((p->attr->scaling&~locked)==UNCONSTRAINED)
			glScalef(p->attr->xscale, p->attr->yscale, p->attr->zscale);

		glTranslatef(-p->attr->cx, -p->attr->cy, -p->attr->cz);

		glCallList(p->attr->idx_objects);

		if ((p->attr->axes&~locked) != NOAXES)
			glCallList(p->attr->idx_bbox);

		if (p->attr->title!=NULL)
			glCallList(p->attr->idx_title);

		glPopMatrix();
		glEndList();
	} else {
		glPushMatrix();
		if ((p->attr->scaling&~locked)==UNCONSTRAINED)
			glScalef(p->attr->xscale, p->attr->yscale, p->attr->zscale);
		glTranslatef(-p->attr->cx, -p->attr->cy, -p->attr->cz);

		compile_objects(p);

		if ((p->attr->axes&~locked) != NOAXES)
			compile_boundingbox(p->attr);

		if (p->attr->title!=NULL)
			compile_title(p->attr);

		glPopMatrix();
	}
}
