#include <stdio.h>
#include <stdlib.h>	/* for free() */
#include <float.h>	/* for DBL_MAX */
#include <ctype.h>
#include <setjmp.h>
#include "dynagraph.h"

/* macro hackery to keep error codes and messages in synch */
enum {
#define MACRO(code,mesg) code,
#include "dataplot.msg"
DPERR_COUNT		/* hides final comma */
#undef MACRO
};

static const char *message[] = {
#define MACRO(code,mesg) mesg,
#include "dataplot.msg"
NULL			/* hides final comma */
#undef MACRO
};


static const char *message_fmt1 = "\terror: file %s -- %s\n";
static const char *message_fmt2 = "\terror: file %s line %d -- %s\n";

#define MAXBUFLEN 128	/* enough to hold the title */

static int lineno;
static char *title;
static char *filename;
static jmp_buf jumpbuf;

/* Scans datafile for surface or curve/tube description data.
   Allocates and fills a mesh matrix or a curve vector.
   In case of mesh, places the grid size in "rows" and "cols"
   In case of curve/tube, places the grid size in "cols"
   Returns 0 on failiure or one of PARAM_PLOT, CURVE_PLOT, TUBE_PLOT 
   on success.

   Datafile is free-formatted, i.e., whitespace (spaces, tabs, newlines)
      between tokens are ignored throughout the file.  Its contents must
      must be as follows:

   1. a header section consisting of
        o optional comments, marked by '#' which may occur anywhere
          (everything beginning with # to the next end-of-line is ignored)
        o one of the four required keywords SURF or CURVE, TUBE or TITLE
	o SURF must be followed by a pair of integers indicating the row
	  and column sizes in surface's grid
	o CURVE and TUBE must be followed an integer indicating the number
	  of data points along the curve
	o An optional TITLE must be followed by a quoted string.
	  Three quote forms are available:  "...", '...', or `...`.
	  A quote may not extend across line boundaries.
   2. a data section consisting of a sequence of coordinate triples,
      each coordinate triple (consisting of a sequence of three floating
      point numbers) describes one point on the surface or on the curve.
      no comments or punctuations are allowed in the data section, otherwise
      the formatting is completely free, as in the header section

Therefore the general format of a SURF datafile is:
TITLE "This is an optional title"
SURF m n
x11 y11 z11    x12 y12 z12   ...   x1n y1n z1n
x21 y21 z21    x22 y22 z22   ...   x2n y2n z2n
...
xm1 ym1 zm1    xm2 ym2 zm2   ...   xmn ymn zmn

The general format of a CURVE datafile is:
CURVE n
x1 y1 z1    x2 y2 z2   ...   xn yn zn

The general format of a TUBE datafile is:
TUBE n
x1 y1 z1    x2 y2 z2   ...   xn yn zn

*/

void skip_comment(FILE *fp)
{
	int c;
	while ((c=fgetc(fp))!=EOF)
		if (c=='\n') {
			ungetc(c, fp);
			return;
		}
}

static void skip_whitespace(FILE *fp)
{
	int c;

	while ((c=fgetc(fp))!=EOF)
		if (c=='\n') {
			lineno++;
			continue;
		} else if (isspace((unsigned char)c))
			continue;
		else if (c=='#') {
			skip_comment(fp);
			continue;
		} else {
			ungetc(c, fp);
			return;
		}
}

static int get_int(FILE *fp)
{
	int n;

	skip_whitespace(fp);

	if (fscanf(fp, "%d", &n) == 1)
		return n;

	printf(message_fmt2, filename, lineno, message[DPERR_BADNUM]);
	longjmp(jumpbuf, 1);
}

static double get_dbl(FILE *fp)
{
	double x;

	skip_whitespace(fp);

	if (fscanf(fp, "%lf", &x) == 1)
		return x;

	printf(message_fmt2, filename, lineno, message[DPERR_BADNUM]);
	longjmp(jumpbuf, 1);
}

static void get_keyword(FILE *fp, char *buf)
{
	int c, i;

	skip_whitespace(fp);

	c = fgetc(fp);

	i = 0;
	if (isalpha((unsigned char)c)) {
		ungetc(c, fp);
		while ((c=fgetc(fp))!=EOF && i<MAXBUFLEN-1)
			if (isalnum((unsigned char)c))
				buf[i++] = c;
			else {
				ungetc(c, fp);
				break;
			}

	}
	buf[i] = '\0';
}

static void get_quoted_string(FILE *fp, char *buf)
{
	int c, quote, i;

	skip_whitespace(fp);

	c = fgetc(fp);
	
	if (c == '"' || c == '\'' || c == '`')
		quote = c;
	else {
		printf(message_fmt2, filename, lineno, message[DPERR_BADSTR]);
		longjmp(jumpbuf, 1);
	}

	i = 0;
	while ((c=fgetc(fp))!=EOF && i<MAXBUFLEN-1 && c!=quote) {
		buf[i++] = c;
		if (c=='\n')
			lineno++;
	}

	buf[i] = '\0';

	if (c!=quote) {
		printf(message_fmt2, filename, lineno, message[DPERR_BADQUOTE]);
		longjmp(jumpbuf, 1);
	}
}

static int readdata(FILE *fp, char *filename,
		Point ***mesh, Point **curve, int *rows, int *cols)
{
	int m=0, n=0;
	unsigned int datatype;
	char buf[MAXBUFLEN];
	int i, j;

	datatype = 0;

	*mesh = NULL;
	*curve = NULL;

	if (setjmp(jumpbuf)!=0) {
		free_mesh(*mesh);
		free(*curve);
		return 0;
	}

	get_keyword(fp, buf);
	
	if (buf[0] == '\0')
		return 0;	/* nothing to read */

	if (strcmp(buf, "TITLE")==0) {
		get_quoted_string(fp, buf);
		testfree(title);
		title = malloc(strlen(buf)+1);
		strcpy(title, buf);
		get_keyword(fp, buf);
	} 

	if (strcmp(buf, "SURF")==0)
		datatype=PARAM_PLOT;
	else if (strcmp(buf, "TUBE")==0)
		datatype=TUBE_PLOT;
	else if (strcmp(buf, "CURVE")==0)
		datatype=CURVE_PLOT;
	else {
		printf(message_fmt2, filename, lineno, message[DPERR_BADCHARS]);
		return 0;
	}

	if (datatype == PARAM_PLOT) {
		m = get_int(fp);
		n = get_int(fp);

		if (m<3 || n<3) {
			printf(message_fmt2, filename, lineno, message[DPERR_BADGRID]);
			return 0;
		}
	} else {	/* CURVE_PLOT or TUBE_PLOT */
		n = get_int(fp);
		if (n<3) {
			printf(message_fmt2, filename, lineno, message[DPERR_BADGRID]);
			return 0;
		}
	}

	/* end of the header section */


	if (datatype==0)
		return 0;

	/* beginning of the data section */
	if (datatype == PARAM_PLOT) {
		*mesh = create_mesh(m, n);
		for (i=0; i<m; i++)
			for (j=0; j<n; j++) {
				(*mesh)[i][j].x = get_dbl(fp);
				(*mesh)[i][j].y = get_dbl(fp);
				(*mesh)[i][j].z = get_dbl(fp);
			}
		*rows = m;
		*cols = n;
		return PARAM_PLOT;
	} else {	/* CURVE_PLOT or TUBE_PLOT */
		*curve = create_curvepoints(n);
		for (j=0; j<n; j++) {
			(*curve)[j].x = get_dbl(fp);
			(*curve)[j].y = get_dbl(fp);
			(*curve)[j].z = get_dbl(fp);
		}
		*cols = n;
		return datatype;
	}
}

Plot3D *make_data_plot(char *infile, Plot3Dattribs *a0)
{
	unsigned int datatype;
	FILE *fp;
	Plot3D *p, *p0;
	Point **mesh;
	Point *curve;
	int m, n;
	double d;


	p0 = p = NULL;
	title = NULL;
	lineno = 1;
	filename = infile;	/* make file name global */

	if ((fp=fopen(filename, "r"))==NULL) {
		printf(message_fmt1, filename, message[DPERR_BADFILE]);
		return NULL;
	}

	while ((datatype=readdata(fp, filename, &mesh, &curve, &m, &n))) {


		if (p==NULL) {
			p0 = p = create_plot3d_structure();
			p0->attr = a0;
		} else {
			p->fp = create_plot3d_structure();
			p->fp->bp = p;
			p = p->fp;
			p->attr = copy_plot3d_attribs_structure(a0);
		}

	
		switch(datatype) {
			case PARAM_PLOT:
				p->type = PARAM_PLOT;
				p->mesh = mesh;
				p->normal = create_mesh(m, n);
				p->attr->m = m;
				p->attr->n = n;
				p->attr->nplots = 1;
				compute_bounds(p);
				compute_normals(p);
				break;
			case CURVE_PLOT:
				p->type = CURVE_PLOT;
				p->curvepoints = curve;
				p->attr->n = n;
				p->attr->nplots = 1;
				compute_curvebounds(p);
				break;
			case TUBE_PLOT:
				if (evalf(a0->tube_points, &d)==0)
					p->attr->m = DEFAULT_TUBE_POINTS;
				else
					p->attr->m = MAX(MIN_TUBE_POINTS, (int)d);

				if (evalf(a0->tube_radius, &d)==0)
					p->attr->tuberadius = DEFAULT_TUBE_RADIUS;
				else
					p->attr->tuberadius = MAX(0, d);

				p->type = TUBE_PLOT;
				p->attr->n = n;
				p->attr->nplots = 1;
				p->mesh = maketube(curve, p->attr->n, p->attr->m, a0->tuberadius);
				free(curve);
				if (p->mesh==NULL) {
					free_plotlist(p0);
					testfree(title);
					fclose(fp);
					return NULL;
				}
				p->normal = create_mesh(p->attr->m, p->attr->n);
				compute_bounds(p);
				compute_normals(p);
				break;
			default:
				free_plotlist(p0);
				testfree(title);
				fclose(fp);
				return NULL;
		}

		a0->xmin = MIN(a0->xmin, p->xmin);
		a0->xmax = MAX(a0->xmax, p->xmax);
		a0->ymin = MIN(a0->ymin, p->ymin);
		a0->ymax = MAX(a0->ymax, p->ymax);
		a0->zmin = MIN(a0->zmin, p->zmin);
		a0->zmax = MAX(a0->zmax, p->zmax);
	}

	fclose(fp);

	/* the section below duplicates a good chunk of the function
	 * evaluate_plots() in parse.c.  Consider eliminating duplication
	 * in future revisions.
	 */
	if (evalf(a0->num_contours, &d)!=0)
		a0->contours = d;

	if (evalf(a0->view_longitude, &d)!=0)
		a0->longitude = d;
	if (evalf(a0->view_colatitude, &d)!=0)
		a0->colatitude = d;

	if (evalf(a0->x_viewmin, &d)!=0)
		a0->xviewmin = d;
	if (evalf(a0->x_viewmax, &d)!=0)
		a0->xviewmax = d;

	if (evalf(a0->y_viewmin, &d)!=0)
		a0->yviewmin = d;
	if (evalf(a0->y_viewmax, &d)!=0)
		a0->yviewmax = d;

	if (evalf(a0->z_viewmin, &d)!=0)
		a0->zviewmin = d;
	if (evalf(a0->z_viewmax, &d)!=0)
		a0->zviewmax = d;

	/* if z-view range is specified, use it */
	if (a0->zviewmin!=DBL_MAX) {
		a0->zmin = a0->zviewmin;
		a0->zmax = a0->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (a0->xviewmin!=DBL_MAX) {
		a0->xmin = a0->xviewmin;
		a0->xmax = a0->xviewmax;
		a0->ymin = a0->yviewmin;
		a0->ymax = a0->yviewmax;
	}

	compute_centroid(a0);

	for (p=p0; p!=NULL; p=p->fp) {
		p->attr->xmin = a0->xmin;
		p->attr->xmax = a0->xmax;
		p->attr->ymin = a0->ymin;
		p->attr->ymax = a0->ymax;
		p->attr->zmin = a0->zmin;
		p->attr->zmax = a0->zmax;
	}

	if (a0->title==NULL)
		a0->title = title;
	else
		testfree(title);

	return p0;
}
