#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dynagraph.h"
#include "gl2ps.h"
#include <GL/glx.h>

Fontdesc		*fonttab;
extern FD_form		*Form;
extern char		*prognam;
extern GLfloat		*fgColor;
extern gl2psContext	*gl2ps;

/* Receives a Fontdesc structure, which should arrive with its fontcode
   field set.
   Computes and fills the xfontinfo, glbase, ascent and descent fields of
   the structure.  If font is not found or out of memory, the glbase
   field is set to zero.
*/
static void make_rasterfont(Fontdesc *fontdesc)
{
	Font id;
	unsigned int first, last;
	char buf1[128], buf2[256];
	char *fallbackfont = "fixed";

	/* build the long font description string */
	strcpy(buf1, "-adobe-");

	/* family */
	if (fontdesc->fontcode & HELVETICA)
		strcat(buf1, "helvetica-");
	else if (fontdesc->fontcode & COURIER)
		strcat(buf1, "courier-");
	else if (fontdesc->fontcode & TIMES)
		strcat(buf1, "times-");
	else if (fontdesc->fontcode & SYMBOL)
		strcat(buf1, "symbol-");
	else {
		printf("%s: we shouldn't be here: %s line %d\n",
				prognam, __FILE__, __LINE__);
		exit(1);
	}

	/* weight */
	if (fontdesc->fontcode & (ROMAN|ITALIC|OBLIQUE))
		strcat(buf1, "medium-");
	else if (fontdesc->fontcode & (BOLD|BOLDOBLIQUE|BOLDITALIC))
		strcat(buf1, "bold-");
	else {
		printf("%s: we shouldn't be here: %s line %d\n",
				prognam, __FILE__, __LINE__);
		exit(1);
	}

	/* slant */
	if (fontdesc->fontcode & (ROMAN|BOLD))
		strcat(buf1, "r-");
	else if (fontdesc->fontcode & (ITALIC|BOLDITALIC))
		strcat(buf1, "i-");
	else if (fontdesc->fontcode & (OBLIQUE|BOLDOBLIQUE))
		strcat(buf1, "o-");
	else {
		printf("%s: we shouldn't be here: %s line %d\n",
				prognam, __FILE__, __LINE__);
		exit(1);
	}

	/* the fontsize is stored in the first 7 bits of fontcode */
	sprintf(buf2, "%s*-*-%d-*-*-*-*-*-*-*", buf1, fontdesc->fontcode&0177);

	fontdesc->xfontinfo = XLoadQueryFont(fl_display, buf2);
	if (fontdesc->xfontinfo==NULL) {
		printf("warning: font '%s' not found, substituting '%s'\n",
					buf2, fallbackfont);
		strcpy(buf2, fallbackfont);
		fontdesc->xfontinfo = XLoadQueryFont(fl_display, buf2);
		if (fontdesc->xfontinfo==NULL) {
			printf("warning: font '%s' not found\n", buf2);
			fontdesc->glbase = 0;
			return;
		}
	}

	id    = fontdesc->xfontinfo->fid;
	first = fontdesc->xfontinfo->min_char_or_byte2;
	last  = fontdesc->xfontinfo->max_char_or_byte2;

	fontdesc->glbase = glGenLists((GLuint)last+1);
	if (fontdesc->glbase==0) {
		printf("warning: unable to allocate memory for font\n");
		return;
	}

	fontdesc->nchars = (GLuint)last+1;
	fontdesc->ascent = fontdesc->xfontinfo->max_bounds.ascent;
	fontdesc->descent = fontdesc->xfontinfo->max_bounds.descent;

	glXUseXFont(id, first, last-first+1, fontdesc->glbase+first);
}

void print_string(const char *s, GLuint base)
{
	if (base==0)
		return;		/* invalid font */

	if (s==NULL)
		return;		/* nothing to print */

	glPushAttrib(GL_LIST_BIT);
	glListBase(base);
	glCallLists(strlen(s), GL_UNSIGNED_BYTE, (GLubyte *)s);
	glPopAttrib();
}

Fontdesc *getfont(unsigned int fontcode)
{
	Fontdesc *f;

	for (f=fonttab; f!=NULL; f=f->fp)
		if (f->fontcode == fontcode)
			return f;

	return NULL;
}

Fontdesc *generate_font(unsigned int fontcode)
{
	Fontdesc *f;

	if ((f=getfont(fontcode))!=NULL)
		return f;

	f = malloc(sizeof(Fontdesc));
	f->fontcode = fontcode;
	make_rasterfont(f);
	f->fp = fonttab;
	fonttab = f;

	return f;
}

/* generate default fonts */
void generate_default_fonts(void)
{
	generate_font(	  DEFAULT_FONTFAMILY
			+ DEFAULT_FONTSTYLE
			+ DEFAULT_FONTSIZE);		/* font */

	generate_font(	  DEFAULT_AXESFONTFAMILY
			+ DEFAULT_AXESFONTSTYLE
			+ DEFAULT_AXESFONTSIZE);	/* axesfont */

	generate_font(	  DEFAULT_LABELFONTFAMILY
			+ DEFAULT_LABELFONTSTYLE
			+ DEFAULT_LABELFONTSIZE);	/* labelfont */

	generate_font(	  DEFAULT_TITLEFONTFAMILY
			+ DEFAULT_TITLEFONTSTYLE
			+ DEFAULT_TITLEFONTSIZE);	/* titlefont */
}

/* Receives fontcode.  Sets the argument `fontname' to the appropriate
   member of a static array which contains the font name a la Postscript.
   Also extracts the font size into the location pointed to by fontsize.
*/
void get_PostscriptFont(unsigned int fontcode, char **fontname, int *fontsize)
{
	/* Important: the order of entries in the following enum and
	   char array must correspond to eachother.
	*/
	enum {
		TimesRoman,
		TimesBold,
		TimesItalic,
		TimesBoldItalic,
		Helvetica,
		HelveticaBold,
		HelveticaOblique,
		HelveticaBoldOblique,
		Courier,
		CourierBold,
		CourierOblique,
		CourierBoldOblique,
		Symbol
	};

	static char *fontnames[] = {
		"Times-Roman",
		"Times-Bold",
		"Times-Italic",
		"Times-BoldItalic",
		"Helvetica",
		"Helvetica-Bold",
		"Helvetica-Oblique",
		"Helvetica-BoldOblique",
		"Courier",
		"Courier-Bold",
		"Courier-Oblique",
		"Courier-BoldOblique",
		"Symbol"
	};

	int lineno;

	if (fontcode & TIMES) {
		if (fontcode & ROMAN)
			*fontname = fontnames[TimesRoman];
		else if (fontcode & BOLD)
			*fontname = fontnames[TimesBold];
		else if (fontcode & ITALIC)
			*fontname = fontnames[TimesItalic];
		else if (fontcode & BOLDITALIC)
			*fontname = fontnames[TimesBoldItalic];
		else {
			lineno = __LINE__;
			goto error;
		}
	} else if (fontcode & HELVETICA) {
		if (fontcode & ROMAN)
			*fontname = fontnames[Helvetica];
		else if (fontcode & BOLD)
			*fontname = fontnames[HelveticaBold];
		else if (fontcode & OBLIQUE)
			*fontname = fontnames[HelveticaOblique];
		else if (fontcode & BOLDOBLIQUE)
			*fontname = fontnames[HelveticaBoldOblique];
		else {
			lineno = __LINE__;
			goto error;
		}
	} else if (fontcode & COURIER) {
		if (fontcode & ROMAN)
			*fontname = fontnames[Courier];
		else if (fontcode & BOLD)
			*fontname = fontnames[CourierBold];
		else if (fontcode & OBLIQUE)
			*fontname = fontnames[CourierOblique];
		else if (fontcode & BOLDOBLIQUE)
			*fontname = fontnames[CourierBoldOblique];
		else {
			lineno = __LINE__;
			goto error;
		}
	} else if (fontcode & SYMBOL)
		*fontname = fontnames[Symbol];

	/* font size is stored in the first 7 bits of fontcode */
	*fontsize = fontcode&0177;

	return;

error:
	printf("%s: error in file %s, line %d\n", prognam, __FILE__, lineno);
	exit(1);
}

void compile_title(Plot3Dattribs *a)
{
	FL_OBJECT *glCanvas = Form->glCanvas;
	double b, u, v, x, y;
	GLuint w, h;
	Fontdesc *fontdesc;
	int fontheight;
	char *p, *t;
	char *psfontname;
	int psfontsize;

	fontdesc = generate_font(a->titlefont&~locked);
	fontheight = fontdesc->ascent + fontdesc->descent;

	if (gl2ps) get_PostscriptFont(a->titlefont&~locked,
				&psfontname, &psfontsize);

	/* the displaylist a->idx_title is allocated in compile_objects() */
	if (!gl2ps) glNewList(a->idx_title, GL_COMPILE);

	glDisable(GL_DEPTH_TEST);
	glPushMatrix();
	glLoadIdentity();

	glPushAttrib(GL_LIGHTING_BIT);
	glDisable(GL_LIGHTING);
	glColor3fv(fgColor);

	b = 1.05 * a->radius;
	w = glCanvas->w;
	h = glCanvas->h;
	u = b * MAX((double)w/h, 1.0);
	v = b * MAX((double)h/w, 1.0);
	y = (glCanvas->h - 2*fontheight) * v / glCanvas->h;

	/* The user can specify a multi-line title, as in:
	   
	         title=`line1\nline2\nline3`
		 
	   The \n combination is receieved literally as a backslash
	   followed by an n.  Here we replace the pair with two
	   consequtive newline characters.  Strtok handles this properly:
	   any number of consequtive separators are equivalent to one
	   separator.  Before altering the title, however, we copy it
	   into a termorary buffer, because the original title string
	   may be needed later, for instance when the user changes the
	   window's background color.
	*/
	t = p = copy_string(a->title);
	
	while ((p=strstr(p, "\\n"))!=NULL) {
		*p++ = '\n';
		*p++ = '\n';
	}

	p = strtok(t, "\n");

	do {
		int textwidth = XTextWidth(fontdesc->xfontinfo, p, strlen(p));

		x = - textwidth * u / glCanvas->w;
		x = MAX(x, -a->radius);

		if (gl2ps) {
			glRasterPos3f(x, y, +3*b);
			gl2psText(p, psfontname, psfontsize);
		} else
			glRasterPos2f(x, y);

		print_string(p, fontdesc->glbase);

		y -= 2.0 * fontheight * v / glCanvas->h;
	} while ((p=strtok(NULL, "\n")) != NULL);

	free(t);

	glPopAttrib();	/* GL_LIGHTING_BIT */

	glPopMatrix();
	glEnable(GL_DEPTH_TEST);
	if (!gl2ps) glEndList();
}
