#include <stdlib.h>
#include <setjmp.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <GL/gl.h>
#include "gl2ps.h"
#include "bsp.h"
#include "dynagraph.h"		/* for DYNAGRAPH_VERSION */

/* DEBUG is set in bsp.h */
#ifdef DEBUG
#define MIN(a,b)	((a)<(b) ? (a) : (b))
#define MAX(a,b)	((a)>(b) ? (a) : (b))
#endif /* DEBUG */

#define DEFAULT_BUFSIZE	1024	/* TEMPORARY FOR TESTNG, NEED SOMETHING BIG */
#define RGB2GRAY(r,g,b)		(20./64*(r) + 32./64*(g) + 12./64*(b))
static GLsizei bufsize = 0;
gl2psContext *gl2ps = NULL;
jmp_buf gl2psJmpEnv;
extern float bspEpsilon;
extern const char gl2ps_prolog[];

void *gl2psMalloc(size_t nbytes)
{
	void *m = malloc(nbytes);
	if (m == NULL && nbytes != 0) {
		fprintf(stderr, "\tout of memory!\n");
		longjmp(gl2psJmpEnv, 1);
	}
	return m;
}

/* reads object data from OpenGL's feedback buffer and creates object */
static bspObj *extract_object(GLfloat *buffer, bspObjType type, short nverts)
{
#ifdef DEBUG
	/* beware that these static values are _not_ reset in repeated calls */
	static int point_no = 1;
	static int line_no = 1;
	static int triangle_no = 1;
#endif /* DEBUG */
	bspObj *obj;
	short i, j, k;

	obj = bspMakeObject(type, nverts);

#ifdef DEBUG
	obj->name = gl2psMalloc(32);
	if (type==bspPointObj)
		sprintf(obj->name, "[P%d]", point_no++);
	else if (type==bspLineObj)
		sprintf(obj->name, "[L%d]", line_no++);
	else if (type==bspTriangleObj)
		sprintf(obj->name, "[T%d]", triangle_no++);
	else
		/* shouldn't be here; bspTextObj is handled in gl2psText() */
		assert(0);
#endif /* DEBUG */

	obj->depth = 0.0;
	for (i=0, k=0; i<nverts; i++) {
		for (j=0; j<3; j++, k++)
			obj->v[i][j] = buffer[k];

		obj->depth += obj->v[i][2];
#ifdef DEBUG
		gl2ps->mindepth = MIN(gl2ps->mindepth, obj->v[i][2]);
		gl2ps->maxdepth = MAX(gl2ps->maxdepth, obj->v[i][2]);
#endif /* DEBUG */

		for (j=0; j<4; j++, k++)
			obj->rgba[i][j] = buffer[k];
	}

	obj->depth /= nverts;

	PRINTF(("extracted object `%s'\n", obj->name));

	return obj;
}

static void parse_feeback_buffer(void)
{
	GLint i;
	GLfloat token;
	bspObj *obj;
	GLfloat pointsize = GL2PS_DEFAULT_POINTSIZE;
	GLfloat linewidth = GL2PS_DEFAULT_LINEWIDTH;

	for (i=0; i<gl2ps->buflen; /* empty */) {
		token = gl2ps->buf[i++];
		switch((int)token) {
			case GL_PASS_THROUGH_TOKEN:
				PRINTF(("GL_PASS_THROUGH_TOKEN\n"));
				if (gl2ps->buf[i] == GL2PS_SET_POINT_SIZE) {
					assert(gl2ps->buf[i+1] == GL_PASS_THROUGH_TOKEN);
					PRINTF(("setting pointsize to %g\n", pointsize));
					pointsize = gl2ps->buf[i+2];
				} else if (gl2ps->buf[i] == GL2PS_SET_LINE_WIDTH) {
					assert(gl2ps->buf[i+1] == GL_PASS_THROUGH_TOKEN);
					PRINTF(("setting linewidth to %g\n", linewidth));
					linewidth = gl2ps->buf[i+2];
				} else {
					fprintf(stderr, "unrecognized token %g in gl2ps->buf\n", gl2ps->buf[i]);
					longjmp(gl2psJmpEnv, 1);
				}
				i += 3;
				break;
			case GL_POINT_TOKEN:
				PRINTF(("GL_POINT_TOKEN\n"));
				obj = extract_object(gl2ps->buf+i, bspPointObj, 1);
				obj->pointsize = pointsize;
				bspObjAdd(gl2ps->rootnode, obj);
				gl2ps->count_before[obj->type]++;
				i += 7;
				break;
			case GL_LINE_RESET_TOKEN:
				PRINTF(("GL_LINE_RESET_TOKEN\n"));
				/* FALL TRHOUGH */
			case GL_LINE_TOKEN:
				PRINTF(("GL_LINE_TOKEN\n"));
				obj = extract_object(gl2ps->buf+i, bspLineObj, 2);
				obj->linewidth = linewidth;
				bspObjAdd(gl2ps->rootnode, obj);
				gl2ps->count_before[obj->type]++;
				i += 2*7;
				break;
			case GL_POLYGON_TOKEN:
				PRINTF(("GL_POLYGON_TOKEN\n"));
				PRINTF(("  n = %g\n", gl2ps->buf[i]));
				if (gl2ps->buf[i]!=3) {
					fprintf(stderr, "received a polygon of %g vertices\n", gl2ps->buf[i]);
					fprintf(stderr, "cannot handle polygons other than triangles!\n");
					longjmp(gl2psJmpEnv, 1);
				}
				i++;
				obj = extract_object(gl2ps->buf+i, bspTriangleObj, 3);
				bspObjAdd(gl2ps->rootnode, obj);
				gl2ps->count_before[obj->type]++;
				i += 3*7;
				break;
			case GL_BITMAP_TOKEN:
				PRINTF(("skipping bitmap token in feedback buffer\n"));
				i += 7;
				break;
			default:
				fprintf(stderr, "unrecognized token %g in gl2ps->buf\n", token);
				longjmp(gl2psJmpEnv, 1);
		}
	}
}

static void write_text_data(bspObj *obj)
{
	short i = 0;

	assert(obj->type == bspTextObj);

#ifdef DEBUG
	fprintf(gl2ps->stream, "%% Text %s\n", obj->name);
#endif /* DEBUG */

	if (gl2ps->shademode == gl2psCOLOR) {
		fprintf(gl2ps->stream, "%.3f %.3f %.3f C ",
				obj->rgba[i][0],
				obj->rgba[i][1],
				obj->rgba[i][2]);
	} else {
		fprintf(gl2ps->stream, "%.3f G ",
				RGB2GRAY(obj->rgba[i][0],
				obj->rgba[i][1],
				obj->rgba[i][2]));
	}

	fprintf(gl2ps->stream, "(%s) %g %g %g /%s S\n",
			obj->text->str, obj->v[i][0], obj->v[i][1],
			obj->text->fontsize, obj->text->fontname);
}

static void write_point_data(bspObj *obj)
{
	short i = 0;

	assert(obj->type == bspPointObj);

#ifdef DEBUG
	fprintf(gl2ps->stream, "%% POINT %s\n", obj->name);
#endif /* DEBUG */


	if (gl2ps->shademode == gl2psCOLOR) {
		fprintf(gl2ps->stream, "%.3f %.3f %.3f C ",
				obj->rgba[i][0],
				obj->rgba[i][1],
				obj->rgba[i][2]);
	} else {
		fprintf(gl2ps->stream, "%.3f G ",
				RGB2GRAY(obj->rgba[i][0],
				obj->rgba[i][1],
				obj->rgba[i][2]));
	}


	fprintf(gl2ps->stream, "%g %g %.2f P\n",
			obj->v[i][0], obj->v[i][1], obj->pointsize/2.0);
}

static void write_line_data(bspObj *obj)
{
	short i;

	assert(obj->type == bspLineObj);

#ifdef DEBUG
	fprintf(gl2ps->stream, "%% LINE %s\n", obj->name);
#endif /* DEBUG */

	if (gl2ps->linewidth != obj->linewidth) {
		fprintf(gl2ps->stream, "%g W ", obj->linewidth);
		gl2ps->linewidth = obj->linewidth;
	}

	if (gl2ps->shademode == gl2psCOLOR) {
		for (i=0; i<2; i++)
			fprintf(gl2ps->stream, "%g %g %.3f %.3f %.3f ",
					obj->v[i][0],
					obj->v[i][1],
					obj->rgba[i][0],
					obj->rgba[i][1],
					obj->rgba[i][2]);
		fprintf(gl2ps->stream, "CL\n");
	} else {
		for (i=0; i<2; i++)
			fprintf(gl2ps->stream, "%g %g %.3f ",
					obj->v[i][0],
					obj->v[i][1],
					RGB2GRAY(
					obj->rgba[i][0],
					obj->rgba[i][1],
					obj->rgba[i][2]));
		fprintf(gl2ps->stream, "GL\n");
	}
}

static void write_triangle_data(bspObj *obj)
{
	short i;

	if (gl2ps->sortmode == gl2psBSP_SORT
			|| gl2ps->sortmode == gl2psSIMPLE_AND_BSP_SORT)
		assert(obj->next==NULL);

#ifdef DEBUG
	fprintf(gl2ps->stream, "%% TRIANGLE %s\n", obj->name);
#endif /* DEBUG */

	if (gl2ps->shademode == gl2psCOLOR) {
		for (i=0; i<3; i++)
			fprintf(gl2ps->stream, "%g %g %.3f %.3f %.3f ",
					obj->v[i][0],
					obj->v[i][1],
					obj->rgba[i][0],
					obj->rgba[i][1],
					obj->rgba[i][2]);
		fprintf(gl2ps->stream, "CT\n");
	} else {
		for (i=0; i<3; i++)
			fprintf(gl2ps->stream, "%g %g %.3f ",
					obj->v[i][0],
					obj->v[i][1],
					RGB2GRAY(obj->rgba[i][0],
					obj->rgba[i][1],
					obj->rgba[i][2]));
		fprintf(gl2ps->stream, "GT\n");
	}
}

static void linked_list_print(bspObj *obj)
{
	for (/* empty */; obj!=NULL; obj=obj->next) {
		switch (obj->type) {
			case bspTextObj:
				write_text_data(obj);
				break;
			case bspPointObj:
				write_point_data(obj);
				break;
			case bspLineObj:
				write_line_data(obj);
				break;
			case bspTriangleObj:
				write_triangle_data(obj);
		}
		gl2ps->count_after[obj->type]++;
		PRINTF(("printed data for %s\n", obj->name));
	}
}

static void bsp_tree_traverse_print(bspNode *node, bspPoint view)
{
	double z;
	
	if (node==NULL)
		return;

	z = bspDotprod(view, (void*)(node->plane)) - node->plane[3];
	PRINTF(("\tlooking at `%s' with dotprod = %g\n", node->obj->name, z));
	if (z>0) {
		bsp_tree_traverse_print(node->left, view);
		linked_list_print(node->obj);
		bsp_tree_traverse_print(node->right, view);
	} else {
		bsp_tree_traverse_print(node->right, view);
		linked_list_print(node->obj);
		bsp_tree_traverse_print(node->left, view);
	}
}

static void print_final_stats(void)
{
	fprintf(stderr, "gl2ps statistics:\n");
	fprintf(stderr, "\taccuracy: %g, smoothness: %g:\n",
			bspEpsilon, gl2ps->smoothness);
	fprintf(stderr, "\tsorting mode: ");
	if (gl2ps->sortmode==gl2psNO_SORT)
		fprintf(stderr, "gl2psNO_SORT\n");
	else if (gl2ps->sortmode==gl2psSIMPLE_SORT)
		fprintf(stderr, "gl2psSIMPLE_SORT\n");
	else if (gl2ps->sortmode==gl2psBSP_SORT)
		fprintf(stderr, "gl2psBSP_SORT\n");
	else if (gl2ps->sortmode==gl2psSIMPLE_AND_BSP_SORT)
		fprintf(stderr, "gl2psSIMPLE_AND_BSP_SORT\n");
	fprintf(stderr, "\tfeedback buffer size: allocated %u bytes, used %u bytes\n",
			bufsize, gl2ps->buflen);
	fprintf(stderr, "\tpoints: %u\n", gl2ps->count_before[bspPointObj]);
	fprintf(stderr, "\ttext: %u\n", gl2ps->count_before[bspTextObj]);
	fprintf(stderr, "\tlines:  before %u, after %u\n",
			gl2ps->count_before[bspLineObj],
			gl2ps->count_after[bspLineObj]);
	fprintf(stderr, "\ttriangles:  before %u, after %u\n",
			gl2ps->count_before[bspTriangleObj],
			gl2ps->count_after[bspTriangleObj]);
	fprintf(stderr, "\toutput file: %s\n", gl2ps->filename);
}

static gl2psContext *make_gl2ps_context(struct gl2psArgs args)
{
	gl2psContext *context = gl2psMalloc(sizeof *context);
	int i, n;

	bufsize = bufsize==0 ? DEFAULT_BUFSIZE : 2*bufsize;
	PRINTF(("trying bufsize = %u\n", bufsize));
	context->buf = gl2psMalloc(bufsize * sizeof *gl2ps->buf);
	context->buflen = 0;
	context->rootnode = bspMakeNode();
#ifdef DEBUG
	context->mindepth = FLT_MAX;
	context->maxdepth = -FLT_MAX;
#endif /* DEBUG */

	n = sizeof(context->count_before) / sizeof(context->count_before[0]);
	for (i=0; i<n; i++) {
		context->count_before[i] = 0;
		context->count_after[i] = 0;
	}

	context->filename = args.file;
	context->sortmode = args.sortmode;
	context->shademode = args.shademode;
	bspEpsilon = args.accuracy;
	context->smoothness = args.smoothness;
	context->linewidth = GL2PS_DEFAULT_LINEWIDTH;
	context->verbose = args.verbose;

	return context;
}

static void free_gl2ps_context(void)
{
	free(gl2ps->buf);
	bspFreeTree(gl2ps->rootnode);
	free(gl2ps);
}

static void cleanup(void)
{
	if (gl2ps == NULL)
		return;
	if (gl2ps->buflen >= 0)
		bufsize = 0;
	if (gl2ps->stream!=NULL)
		fclose(gl2ps->stream);
	free_gl2ps_context();
	gl2ps = NULL;
}

void gl2psBegin(struct gl2psArgs args)
{
	FILE *stream;

	assert(gl2ps==NULL);

	if ((stream = fopen(args.file, "w"))==NULL) {
		fprintf(stderr, "\tunable to open file %s for writing\n",
				args.file);
		return;
	}

	if (setjmp(gl2psJmpEnv)!=0) {
		cleanup();
		return;
	}

	gl2ps = make_gl2ps_context(args);
	gl2ps->stream = stream;

	/* start feedback mode */
	glFeedbackBuffer(bufsize, GL_3D_COLOR, gl2ps->buf);
	(void)glRenderMode(GL_FEEDBACK);
}

GLint gl2psEnd(void)
{
	GLint buflen;
	GLint viewport[4];
	GLfloat bg[4];
	bspPoint view;
	time_t now;

	if (gl2ps == NULL) {
		fprintf(stderr, "nothing to do in gl2psEnd()\n");
		return 0;
	}


	/* switch to render mode and get length of feedback buffer */
	buflen = gl2ps->buflen = glRenderMode(GL_RENDER);

	/* inadequate bufsize results in negative buflen */
	if (buflen <0) {
		cleanup();
		return buflen;
	}

	if (setjmp(gl2psJmpEnv)!=0) {
		cleanup();
		return buflen;
	}

	parse_feeback_buffer();

	if (gl2ps->rootnode == NULL || gl2ps->rootnode->obj == NULL) {
		fprintf(stderr, "nothing to do in gl2psEnd()\n");
		cleanup();
		return 0;
	}

	if (gl2ps->sortmode == gl2psSIMPLE_SORT
			|| gl2ps->sortmode == gl2psSIMPLE_AND_BSP_SORT)
		bspSimpleSort(gl2ps->rootnode);

	if (gl2ps->sortmode == gl2psBSP_SORT
			|| gl2ps->sortmode == gl2psSIMPLE_AND_BSP_SORT)
		if (gl2ps->rootnode->obj->type == bspTriangleObj)
			bspProduceTree(gl2ps->rootnode);

	glGetIntegerv(GL_VIEWPORT, viewport);
	glGetFloatv(GL_COLOR_CLEAR_VALUE, bg);

	time(&now);
	
	fprintf(gl2ps->stream, "%%!PS-Adobe-3.0 EPSF-3.0\n");
	fprintf(gl2ps->stream, "%%%%Title: %s\n", gl2ps->filename);
	fprintf(gl2ps->stream, "%%%%Creator: Dynagraph %s -- Rouben Rostamian <rostamian@umbc.edu>\n", DYNAGRAPH_VERSION);
	fprintf(gl2ps->stream, "%%%%CreationDate: %s", ctime(&now));
	fprintf(gl2ps->stream, "%%%%Pages: 1\n");
	fprintf(gl2ps->stream, "%%%%BoundingBox: %d %d %d %d\n",
		viewport[0], viewport[1], viewport[2], viewport[3]);
	fprintf(gl2ps->stream, "%%%%EndComments\n");
	fprintf(gl2ps->stream, "%%%%BeginProlog\n");
	fprintf(gl2ps->stream, "/gl2psdict 64 dict def\n");
	fprintf(gl2ps->stream, "gl2psdict begin\n");
	fprintf(gl2ps->stream, "/smoothness %g def\n", gl2ps->smoothness);
	fprintf(gl2ps->stream, "%s\n", gl2ps_prolog);
	fprintf(gl2ps->stream, "%%%%Page 1 1\n");
	fprintf(gl2ps->stream, "gsave\n");
	fprintf(gl2ps->stream, "gl2psdict begin\n");
	fprintf(gl2ps->stream, "1 setlinecap 1 setlinejoin 1 setlinewidth\n");
	fprintf(gl2ps->stream, "%g %g %g C\n", bg[0], bg[1], bg[2]);
	fprintf(gl2ps->stream,
		"newpath %d %d moveto "
		"%d %d lineto %d %d lineto %d %d lineto ",
		viewport[0], viewport[1],
		viewport[2], viewport[1],
		viewport[2], viewport[3],
		viewport[0], viewport[3]);
	fprintf(gl2ps->stream, "closepath fill\n");

#ifdef DEBUG
	fprintf(stderr, "min depth = %g\n", gl2ps->mindepth);
	fprintf(stderr, "max depth = %g\n", gl2ps->maxdepth);
#endif /* DEBUG */

	/* we want to look at the scene from "infinity", so we need something
	   large for view[2].  I think the depth of the scene is always
	   between 0 and 1, so 100.0 is a good approximation for infinity.
	*/
	view[0] = view[1] = 0;
	view[2] = -100.0;

	switch (gl2ps->sortmode) {
		case gl2psBSP_SORT:
		case gl2psSIMPLE_AND_BSP_SORT:
			if (gl2ps->rootnode->obj->type == bspTriangleObj) {
				bsp_tree_traverse_print(gl2ps->rootnode, view);
				break;
			}
			/* FALL TRHOUGH */
		default:
			linked_list_print(gl2ps->rootnode->obj);
	}

	fprintf(gl2ps->stream, "end     %% close gl2ps dict\n");
	fprintf(gl2ps->stream, "grestore\n");
	fprintf(gl2ps->stream, "showpage\n");
	fprintf(gl2ps->stream, "%%EOF\n");

	if (gl2ps->verbose)
		print_final_stats();

	cleanup();

	return buflen;
}

GLvoid gl2psPointSize(GLfloat size)
{
	if(gl2ps == NULL)
		return;


	glPassThrough(GL2PS_SET_POINT_SIZE);
	glPassThrough(size);
}

GLvoid gl2psLineWidth(GLfloat width)
{
	if(gl2ps == NULL)
		return;

	glPassThrough(GL2PS_SET_LINE_WIDTH);

	glPassThrough(width);
}

/* the following function is essentiall an exact copy of the function of
   the same name in Christophe Geuzaine's gl2ps library.
*/
GLvoid gl2psText(const char *str, const char *fontname, GLint fontsize)
{
#ifdef DEBUG
	/* beware that the static text_no is _not_ reset in repeated calls */
	static int text_no = 1;
#endif /* DEBUG */
	bspObj *obj;
	GLfloat pos[4];
	GLboolean valid;

	if (gl2ps == NULL)
		return;

	if (str == NULL)
		return;

	glGetBooleanv(GL_CURRENT_RASTER_POSITION_VALID, &valid);

	if (!valid)
		return;

	glGetFloatv(GL_CURRENT_RASTER_POSITION, pos);

	obj = bspMakeObject(bspTextObj, 1);

#ifdef DEBUG
	obj->name = gl2psMalloc(32);
	sprintf(obj->name, "[W%d]", text_no++);
#endif /* DEBUG */

	obj->v[0][0] = pos[0];
	obj->v[0][1] = pos[1];
	obj->v[0][2] = pos[2];

	obj->depth = pos[2];

	glGetFloatv(GL_CURRENT_RASTER_COLOR, obj->rgba[0]);

	obj->text = gl2psMalloc(sizeof(Text));
	obj->text->str = gl2psMalloc(strlen(str)+1);
	strcpy(obj->text->str, str);
	obj->text->fontname = gl2psMalloc(strlen(fontname)+1);
	strcpy(obj->text->fontname, fontname);
	obj->text->fontsize = fontsize;

	PRINTF(("extracted object `%s'\n", obj->name));

	bspObjAdd(gl2ps->rootnode, obj);
	gl2ps->count_before[obj->type]++;
}
