#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#define SH_CMD_PAGE "test -f ${FILE} -a -r ${FILE} && ${PAGER:-more} ${FILE} \
	|| echo Unable to read file \\\"${FILE}\\\""

#define NOMATCH		-1
#define MULTIMATCH	-2

struct helptable {
	char *s;	/* help request string typed by the user */
	char *t;	/* file name containing help on that topic */
};

static struct helptable topics[] = {
	{ "sample-session",	"sample-session.hlp" },
	{ "examples",		"examples.hlp" },

	{ "plot",		"plotting_surfaces.hlp" },
	{ "spacecurve",		"plotting_curves.hlp" },
	{ "tubeplot",		"plotting_tubes.hlp" },
	{ "dataplot",		"plotting_data.hlp" },
	{ "polygonplot",	"plotting_polygons.hlp" },
	{ "textplot",		"plotting_text.hlp" },
	{ "display",		"merging.hlp" },
	{ "options",		"options.hlp" },

	{ "cylindrical-coords",	"other_coords.hlp" },
	{ "spherical-coords",	"other_coords.hlp" },
	{ "surfaces-of-revolution", "surfaces_of_revolution.hlp" },

	{ "read",		"scripts.hlp" },
	{ "savegraph",		"save_image.hlp" },

	{ "",			"0-index.hlp" },
	{ "?",			"0-index.hlp" },
	{ "help",		"0-index.hlp" },
	{ "restart",		"restart-exit.hlp" },
	{ "quit",		"restart-exit.hlp" },
	{ "exit",		"restart-exit.hlp" },
	{ "!",			"shell_escape.hlp" },
	{ "dump",		"assignments.hlp" },

	{ "statements",		"statements.hlp" },
	{ "assignments",	"assignments.hlp" },
	{ "quoting",		"quoting.hlp" },
	{ "built-ins",		"built-ins.hlp" },
	{ "rotating",		"globe.hlp" },
	{ "lighting",		"lighting.hlp" },
	{ "frames/seconds",	"fps.hlp" },

	{ "license",		"license.hlp" },
	{ "where-to-get",	"where-to-get.hlp" },
	{ NULL }
};

/* Checks if the argument string matches the beginning of
   any of the strings in the first column of topics[].  If no match,
   returns NOMATCH, if multiple matches, returns MULTIMATCH, both of
   which are negative numbers.  Otherwise returns the index into the
   array topics[] of where the unique match was found.

   A delicate issue is what to do when the topics[] array declares
   something like "example" and "examples", as two distinct entries.
   In the routine below we do some maneuvering so that we won't reject
   the string "example" as being `ambiguous'.  In truth, there is some
   ambiguity here and it is probably not a good decision to allow both
   "example" and "examples" as entries, but here we do the best we can.

   This function is extracted from my pcl.c program.  -- RR
*/
static int get_table_index(const char *s)
{
	struct helptable *p;
	int matchcount, exactmatchcount;
	int i, idx1, idx2;

	matchcount = 0;
	exactmatchcount = 0;
	idx1 = idx2 = -1;	/* not required; just silencing -Wall */

	for (p=topics, i=0; p->s!=NULL; p++, i++)
		if (strcmp(p->s, s)==0) {
			exactmatchcount++;
			idx1 = i;
		} else if (strncmp(p->s, s, strlen(s))==0) {
			matchcount++;
			idx2 = i;
		}

	/* exactmatchcount should be either 0 or 1.  Otherwise it
	   means that we have duplicate entries in the first
	   column of the table.
	*/
	assert(exactmatchcount==0 || exactmatchcount==1);

	if (exactmatchcount==1)
		return idx1;

	/* if we arrive here, exactmatchcount is 0 */
	if (matchcount==0)
		return NOMATCH;
	else if (matchcount==1)
		return idx2;
	else /* matchcount>1 */
		return MULTIMATCH;
}

/* This function is invoked when the user types one of "?", "?topic",
   "help", "help topic".  The argument `s' is what the user typed,
   but it has been trimmed of leading and trailing whitespace.
   There may be whitespace preceeding `topic'.
*/
void show_help(char *s)
{
	char *helpdir;
	char *buf;
	char *t;
	int i;

	if (s[0]=='?')
		s += 1;		/* skip over `?' */
	else
		s += 4;		/* skip over `help' */

	while (isspace(*s))
		s++;		/* skip over initial whitespace */

	if ((t=strpbrk(s, ";: "))!=NULL)
		*t = '\0';	/* remove unwanted trailing characters */

	if ((i=get_table_index(s))==NOMATCH) {
		printf("\tNo help available on `%s'.\n", s);
		printf("\tType `?' for a list of all help topics.\n");
		return;
	} else if (i==MULTIMATCH) {
		printf("\tAmbiguous request `%s'.\n", s);
		printf("\tType `?' for a list of all help topics.\n");
		return;
	}

	/* 1. The environment variable DYNAGRAPH_HELPDIR overrides the
	      built-in HELPDIR.
	 
	   2. The built-in HELPDIR is passed to this file through make
	      and the compiler's -D flag.
	 */
	if ((helpdir=getenv("DYNAGRAPH_HELPDIR"))==NULL)
		helpdir = HELPDIR;

	/* 100 char padding to accommodate "FILE=" and "/the_t_string.hlp; " */
	if ((buf=malloc(100+strlen(helpdir)+strlen(SH_CMD_PAGE)))==NULL) {
		printf("\tOut of memory!\n");
		return;
	}

	sprintf(buf, "FILE=%s/%s; %s", helpdir, topics[i].t, SH_CMD_PAGE);

	system(buf);
	free(buf);
}
