#include <sys/time.h>	/* for gettimeofday() */
#include <math.h>	/* for sqrt() */
#include <forms.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include "formdefs.h"
#include "dynagraph.h"

#define deg_to_rad(x)	(M_PI/180.0*x)

extern FD_form		*Form;
static GLfloat		autorotvec[4];

static struct timeval capture_time(void)
{
	static struct timeval t;
	gettimeofday(&t, NULL);
	return t;
}

static double elapsed_time(struct timeval t1)
{
	struct timeval t2;

	gettimeofday(&t2, NULL);
	if (t2.tv_usec < t1.tv_usec) {
		t2.tv_usec += 1000000;
		t2.tv_sec  -= 1;
	}
	return (t2.tv_sec - t1.tv_sec) + 0.000001*(t2.tv_usec - t1.tv_usec);
}

/* Routine to compute the axis and angle of rotation of a "glass sphere"
   inverted on the canvas, under the action of the mouse.
 
   The mouse moves from the point (x,y) on canvas by the amount (dx, dy).
   This causes the glass sphere to rotate about an axis (w1,w2,w3) by an
   angle "rot".  The axis is computed with respect to the coordinates
   attached to the sphere, not the world frame.
*/
static void find_axis_of_rotation(
		int x, int y, int dx, int dy,
		GLfloat *rot, GLfloat *w1, GLfloat *w2, GLfloat *w3)
{
	FL_OBJECT *glCanvas = Form->glCanvas;
	GLfloat m[16];
	float z, dz;
	float R2, r2;
	FL_COORD w, h;
	float a1, a2, a3;

	w = glCanvas->w;
	h = glCanvas->h;

	R2 = (w*w+h*h)/4;	/* sqrt(R2) is window's half-diagonal */

	/* z determines the mouse contact point (x,y,z) on the
	 * "invisible glass" sphere (of radius sqrt(R2).
	*/
	r2 = x*x + y*y;
	z = sqrt(R2-r2);

	/* The motion of the mouse produces a vector (dx,dy,dz).
	 * The first two components are read off the screen.  The
	 * third component is computed assuming that the vector
	 * (dx,dy,dz) is tangent to the sphere, thus orthogonal
	 * to the vector (x,y,z).
	*/
	dz = -(x*dx+y*dy)/z;

	/* the action of the displacement vector (dx,dy,dz) at the
	 * contact point (x,y,z) rotates the virtual "glass sphere" about
	 * an axis (a1,a2,a3) which is computed as the cross-product
	 * of the vectors (x,y,z) and (dx,dy,dz).
	*/
	a1 = y*dz-z*dy;
	a2 = z*dx-x*dz;
	a3 = x*dy-y*dx;

	/* the rotation angle equals displacement over the radius.  */
	*rot = (180/M_PI)*sqrt(dx*dx+dy*dy+dz*dz)/sqrt(R2);

	/* The components (a1,a2,a3) of the axis of rotation w are
	 * with respect to the coordiantes fixed to the screen.
	 * We convert this to the model's local coordinates and
	 * arrive at the rotation axis (w1, w2, w3) which
	 * is computed as M^T a, where M is the ModelView matrix.
	*/
	glGetFloatv(GL_MODELVIEW_MATRIX,m);
	*w1 = m[0]*a1 + m[1]*a2 + m[2]*a3;
	*w2 = m[4]*a1 + m[5]*a2 + m[6]*a3;
	*w3 = m[8]*a1 + m[9]*a2 + m[10]*a3;
}

void autorotate(void)
{
	glRotatef(autorotvec[0], autorotvec[1], autorotvec[2], autorotvec[3]);
	display();
}

static void setup_viewport(Plot3D *p)
{
	FL_OBJECT *glCanvas;
	double radius, b, u, v;
	GLuint w, h;

	if (Form!=NULL) {
		glCanvas = Form->glCanvas;
		w = glCanvas->w;
		h = glCanvas->h;
	} else {	/* no form is shown */
		glCanvas = NULL;
		w = h = 448;	/* default size of glCanvas, as defined in formdefs.c */
	}

	radius = p->attr->radius;
	b = 1.05 * radius;
	u = b * MAX((double)w/h, 1.0);
	v = b * MAX((double)h/w, 1.0);

	glViewport(0,0, w, h);

	glPushMatrix();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-u, u, -v, v, MAXZOOM*b, MAXZOOM*3*b);
	glTranslatef(0, 0, -MAXZOOM*2*b);
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}

/* sets the viewing position as if looking towards the origin
   from a poing on  the ray with (longitude, colatitude) angles
   in the spherical coordiantes */
void reset_glCanvas(Plot3D *p)
{
	double longitude = p->attr->longitude;
	double colatitude = p->attr->colatitude;

	setup_viewport(p);
	glLoadIdentity();
	glRotatef(-(90+longitude), 0, 0, 1);
	glRotatef(colatitude,
			 sin(deg_to_rad(longitude)),
			-cos(deg_to_rad(longitude)), 0);
}

static void write_speed_data(double secsperframe)
{
	char buf[32];
	double val;

	fl_get_button(Form->framecounterbutton)
		? sprintf(buf, "%.1f", val = 1.0/secsperframe)
		: sprintf(buf, "%.3f", val = secsperframe);
	fl_set_object_label(Form->framespeed_box, buf);

	/* Saved value is used by the invert_speed_data() callback function.
	   This will cause 32 bit long-int overflow if val is larger than
	   ~2000.  But realistically, neither frames/sec nor secs/frame 
	   approach 2000.  (However, with hardware rendering we /do/ get
	   erroneous large numbers, in which case the displayed information
	   is garbage.  See comment preceeding display().
	*/
	Form->framespeed_box->u_ldata = (long)(1.0e6 * val);
}

/* The elapsed time comes /much/ too small with hardware rendering, because
   the call glCallList() returns immediately in that case.  On an E&S graphics
   card I recorded speeds of 18,000 draws per second, while the actual value
   is more like 50 draws per second.  I think a call to glFinish() after
   glXSwapBuffers() will force waiting for the completion of
   drawing, but I don't do that here, because the resulting round-trip
   communication between the client and server can take a toll on performance.
*/
void display(void)
{
	FL_OBJECT *glCanvas = Form->glCanvas;
	Plot3D *p = glCanvas->u_vdata;
	struct timeval t = capture_time();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glCallList(p->attr->idx_overall);
	glXSwapBuffers(fl_display, fl_get_canvas_id(glCanvas));
	glFinish();	/* 	TEMPORARY */
	write_speed_data(elapsed_time(t));
}

/* computes total number of nodes displayed in a compound plot */
static int nnodes(Plot3D *p)
{
	int nnodes = 0;
	Plot3D *q;

	for (q=p; q!=NULL; q=q->fp) {
		if (q->type==CURVE_PLOT)
			nnodes += q->attr->n;	/* a spacecurve */
		else
			nnodes += q->attr->m * q->attr->n;
	}
	return nnodes;
}

void install_plot(Plot3D *p)
{
	FL_OBJECT *glCanvas = Form->glCanvas;
	char buf[16];

	if (glCanvas->u_vdata==p)
		return;		/* p is already installed */

	free_plotlist(glCanvas->u_vdata);

	glCanvas->u_vdata = p;
	reset_glCanvas(p);

	sprintf(buf, "%d nodes at", nnodes(p));
	fl_set_object_label(Form->nodecount_box, buf);

	set_menu_options();

	if ((p->attr->scaling&~locked)==CONSTRAINED)
		fl_set_button(Form->constrained_scaling_button, 1);
	else
		fl_set_button(Form->constrained_scaling_button, 0);

	if ( (p->attr->axes&~locked) != NOAXES)
		fl_set_button(Form->toggle_box_button, 1);
	else
		fl_set_button(Form->toggle_box_button, 0);

	/* reset the autorotate vector (is this a good thing?) */
	autorotvec[0] = 5.0;	/* angle increments (degrees) */
	autorotvec[1] = 0.0;	/* x commonent */
	autorotvec[2] = 0.0;	/* y component */
	autorotvec[3] = 1.0;	/* z component */
}

/* Note:  Just a plain glXSwapBuffers() suffices with MESA.  But with
   Xi Graphics' OpenGL we need to display().  Ideally a preprocessor
   macro should do one or the other, but I don't know of a suitable macro.
*/
int canvas_expose(FL_OBJECT *obj, Window win, int w, int h, XEvent *xev, void *ud)
{
	FL_OBJECT *glCanvas = Form->glCanvas;
	/* glXSwapBuffers(fl_display, fl_get_canvas_id(glCanvas)); */
	display();	/* TEMPORARY */
	return 0;
}

int canvas_resize(FL_OBJECT *obj, Window win, int w, int h, XEvent *xev, void *ud)
{
	FL_OBJECT *glCanvas = Form->glCanvas;

	setup_viewport(glCanvas->u_vdata);
	display();
	return 0;
}

int grabmove(FL_OBJECT *obj, Window win, int w, int h, XEvent *xev, void *ud)
{
	static int prevx, prevy, downx, downy;
	int x, y, dx, dy;
	GLfloat rot, w1, w2, w3;
	static struct timeval t;

	/* If outside the canvas, don't do anything, except if it is a
	   ButtonRelease, in which case resume idle callback.
	   Thanks to Rob Fletcher <rpf1@york.ac.uk>
	*/
	if (xev->xbutton.x < 0 || xev->xbutton.x >= w
			|| xev->xbutton.y < 0 || xev->xbutton.y >= h) {
		if (xev->type==ButtonRelease)
			fl_set_idle_callback(idle_callback, NULL);
		return 0;
	}

	switch(xev->type) {

	case ButtonPress:
		fl_set_idle_callback(NULL, NULL);	/* suspend autorotate */
		downx = prevx = xev->xbutton.x - w/2;
		downy = prevy = h/2 - xev->xbutton.y;
		break;

	case MotionNotify:

		/* if the button is released, we don't want to move anymore */
		if ( (xev->xbutton.state&Button1Mask)==0 )
			break;
		
		/* x and y are the screen coordinates of the mouse, measured
		 * with respect to the center of the canvas.
		*/
		x = xev->xbutton.x - w/2;
		y = h/2 - xev->xbutton.y;

		dx = x - prevx;
		dy = y - prevy;

		/* save the position of the mouse to compute dx and dy
		 * in the next round.
		*/
		prevx = x;
		prevy = y;

		find_axis_of_rotation(x, y, dx, dy, &rot, &w1, &w2, &w3);

		/* Now rotate and re-draw.  We magnify rot by some factor
		   to give the impression of a more agile mouse.  */
		glRotatef(4*rot, w1, w2, w3);
		display();
		t = capture_time();
		break;

	case ButtonRelease:

		if ( (elapsed_time(t) < 0.5 ) &&
				( xev->xbutton.x != downx || xev->xbutton.y != downy ) ) {

			x = xev->xbutton.x - w/2;
			y = h/2 - xev->xbutton.y;
			dx = x - downx;
			dy = y - downy;

			find_axis_of_rotation(x, y, dx, dy, &rot, &w1, &w2, &w3);

			autorotvec[0] = rot;
			autorotvec[1] = w1;
			autorotvec[2] = w2;
			autorotvec[3] = w3;

			fl_set_idle_callback(idle_callback, NULL);
		}
		else {	/* stop autorot if button was held still for > 0.5sec */
			fl_set_button(Form->autorot_button, 0);
			fl_set_idle_callback(idle_callback, NULL);
		}

		break;
	}

	return 0;
}
