#include <stdio.h>
#include <math.h>	/* for sqrt() */
#include <float.h>	/* for DBL_EPSILON */
#include "dynagraph.h"
#include "parse.h"

/* #define DEBUG_TRIANGLES */

#ifdef DEBUG_TRIANGLES
void print_triangle_data(Plot3D *p);
#endif /* DEBUG_TRIANGLES */

/* DBL_EPSILON is defined in double.h as the difference between 1.0 and
   the minimum double greater than 1.0.  It equals 2.2204460492503131e-16
   on linux.
*/
#define EPS	DBL_EPSILON

/* We use this routine to compute normals at nodes where the more
 * accurate routine compute_normals() below fails.  This routine, which
 * is less accurate but more robust, computes the normal vector at a
 * vertex by summing the normals on up to 8 adjoining triangles, as
 * shown in the figure below:
 * 
 * +---+---+
 * |\  |  /|
 * | \ | / |
 * |  \|/  |
 * |---\---|
 * |  /|\  |
 * | / | \ |
 * |/  |  \|
 * +---+---+
 * 
 * The routine receives the index (i,j) of the central vertex and
 * generates the coordinates of the 8 surrounding vertices.  Triangles
 * which fall outside the main domain are ignored.  On the remaining
 * triangles, normals are computed using cross-products and added
 * to obtain an approximate normal (nx, ny, nz) at the central vertex.
*/
static void compute_alt_normal( Plot3D *p, int i, int j, double *nx, double *ny, double *nz)
{
	int idx[9][2] = {
		{1,0}, {1,1}, {0,1}, {-1,1}, {-1,0}, {-1,-1}, {0,-1}, {1,-1}, {1,0}
	};
	int k, i1, i2, j1, j2;
	double a1, a2, a3, b1, b2, b3;
	int n = p->attr->n;
	int m = p->attr->m;

	*nx = *ny = *nz = 0;
	for (k=0; k<8; k++) {
		i1 = i + idx[k][0];
		j1 = j + idx[k][1];
		i2 = i + idx[k+1][0];
		j2 = j + idx[k+1][1];

		if ( i1<0 || i1>=m || j1<0 || j1>=n || i2<0 || i2>=m || j2<0 || j2>=n )
			continue;	/* skip triangles outside the domain */

		if (IS_SINGULAR(p->mesh[i1][j1]) || IS_SINGULAR(p->mesh[i2][j2]))
			continue;

		a1 = p->mesh[i1][j1].x - p->mesh[i][j].x;
		a2 = p->mesh[i1][j1].y - p->mesh[i][j].y;
		a3 = p->mesh[i1][j1].z - p->mesh[i][j].z;
		
		b1 = p->mesh[i2][j2].x - p->mesh[i1][j1].x;
		b2 = p->mesh[i2][j2].y - p->mesh[i1][j1].y;
		b3 = p->mesh[i2][j2].z - p->mesh[i1][j1].z;

		*nx += a2*b3 - a3*b2;
		*ny += a3*b1 - a1*b3;
		*nz += a1*b2 - a2*b1;
	}
}

/* We compute normals to the surface using three-point differencing
 * both in the interior points and at the edges.  Most of the time the
 * scheme works well, however when neighboring nodes coalease, as in the
 * north pole of a sphere, the computation results in normal vectors of
 * length zero.  In such cases we revert to the alternate method in the
 * routine compute_alt_normal() above.  The alternate routine is more
 * robust, but less accurate.  If all normals are computed through the
 * alternate routine, visible seams develop along the lines where a
 * surface closes onto itself, as in a sphere.
*/
void compute_normals(Plot3D *p)
{
	int i, j;
	double nx, ny, nz, len;
	double dxdu, dydu, dzdu, dxdv, dydv, dzdv;
	int n = p->attr->n;
	int m = p->attr->m;
	Point **v = p->mesh;
	short bad_vertex = 0;

	for (i=0; i<m; i++) {
		for (j=0; j<n; j++) {

			if (IS_SINGULAR(v[i][j])) {
				p->normal[i][j].x = 1.0;	/* won't be used */
				p->normal[i][j].y = 0;
				p->normal[i][j].z = 0;
				continue;
			}

			if ( j==0 ) {
				dxdu = 4*v[i][j+1].x - 3*v[i][j].x - v[i][j+2].x;
				dydu = 4*v[i][j+1].y - 3*v[i][j].y - v[i][j+2].y;
				dzdu = 4*v[i][j+1].z - 3*v[i][j].z - v[i][j+2].z;
			} else if ( j==n-1) {
				dxdu = -(4*v[i][j-1].x - 3*v[i][j].x - v[i][j-2].x);
				dydu = -(4*v[i][j-1].y - 3*v[i][j].y - v[i][j-2].y);
				dzdu = -(4*v[i][j-1].z - 3*v[i][j].z - v[i][j-2].z);
			} else {
				dxdu = v[i][j+1].x - v[i][j-1].x;
				dydu = v[i][j+1].y - v[i][j-1].y;
				dzdu = v[i][j+1].z - v[i][j-1].z;
			}

			
			if ( i==0 ) {
				dxdv = 4*v[i+1][j].x - 3*v[i][j].x - v[i+2][j].x;
				dydv = 4*v[i+1][j].y - 3*v[i][j].y - v[i+2][j].y;
				dzdv = 4*v[i+1][j].z - 3*v[i][j].z - v[i+2][j].z;
			} else if ( i==m-1) {
				dxdv = -(4*v[i-1][j].x - 3*v[i][j].x - v[i-2][j].x);
				dydv = -(4*v[i-1][j].y - 3*v[i][j].y - v[i-2][j].y);
				dzdv = -(4*v[i-1][j].z - 3*v[i][j].z - v[i-2][j].z);
			} else {
				dxdv = v[i+1][j].x - v[i-1][j].x;
				dydv = v[i+1][j].y - v[i-1][j].y;
				dzdv = v[i+1][j].z - v[i-1][j].z;
			}

			/* -ve signs because in the matrix mesh[i][j] the row
			   index increases downward while the y axis points upward. */
			nx = -(dydu*dzdv - dzdu*dydv);
			ny = -(dzdu*dxdv - dxdu*dzdv);
			nz = -(dxdu*dydv - dydu*dxdv);

			/* if the normal is almost zero, use the alternate method */
			if ( (len = sqrt(nx*nx+ny*ny+nz*nz)) < EPS ) {

				compute_alt_normal(p, i, j, &nx, &ny, &nz);

				if ( (len = sqrt(nx*nx+ny*ny+nz*nz)) < EPS )
					bad_vertex++;
			}

			p->normal[i][j].x = nx/len;
			p->normal[i][j].y = ny/len;
			p->normal[i][j].z = nz/len;
		}
	}

	if (bad_vertex > 0)
		printf("\twarning: couldn't get rid of degenerate normals\n");

#ifdef DEBUG_TRIANGLES
	print_triangle_data(p);
#endif /* DEBUG_TRIANGLES */

}


#ifdef DEBUG_TRIANGLES

/* triangle data */
typedef struct {
	Point A, B, C;			/* vertices */
	double a, b, c;			/* side lengths */
	double area;			/* area */
	double alpha, beta, gamma;	/* angles */
	double minangle;		/* smallest angle */
} Triangle;

/* angle opposite side a */
double angle(double a, double b, double c)
{
	if (a<1e-10||b<1e-10||c<1e-10)
		printf("Warning: degenerate triangle: a=%g, b=%g, c=%g!\n",
				a, b, c);
	return acos((b*b + c*c - a*a)/(2*b*c));
}

/* distance between two points */
static double dist(Point A, Point B)
{
	double x, y, z;

	x = B.x - A.x;
	y = B.y - A.y;
	z = B.z - A.z;

	return sqrt(x*x + y*y + z*z);
}

void solve_triangle(Triangle *T)
{
	double s;
	T->a = dist(T->B, T->C);
	T->b = dist(T->C, T->A);
	T->c = dist(T->A, T->B);
	s = (T->a+T->b+T->c)/2;
	T->area = sqrt(s*(s-T->a)*(s-T->b)*(s-T->c));
	T->alpha = angle(T->a,T->b,T->c);
	T->beta  = angle(T->b,T->c,T->a);
	T->gamma = angle(T->c,T->a,T->b);
	T->minangle = MIN(T->alpha,MIN(T->beta,T->gamma));
}

/* prints info about the worst-case triangles in the mesh */
void print_triangle_data(Plot3D *p)
{
	int iminarea, jminarea;
	double minarea = 1000;
	int imaxarea, jmaxarea;
	double maxarea = 0;
	int iminangle, jminangle;
	double minangle = 4;
	Triangle T;
	int i, j;
	int count = 0;

	for (i=0; i<p->attr->m-1; i++)
		for (j=0; j<p->attr->n-1; j++) {

			T.A = p->mesh[i][j];
			T.B = p->mesh[i+1][j];
			T.C = p->mesh[i][j+1];
			solve_triangle(&T);
			count++;

			if (T.area<minarea) {
				minarea = T.area;
				iminarea = i;
				jminarea = j;
			}

			if (T.area>maxarea) {
				maxarea = T.area;
				imaxarea = i;
				jmaxarea = j;
			}

			if (T.minangle<minangle) {
				minangle = T.minangle;
				iminangle = i;
				jminangle = j;
			}

			T.A = p->mesh[i][j+1];
			T.B = p->mesh[i+1][j];
			T.C = p->mesh[i+1][j+1];
			solve_triangle(&T);
			count++;

			if (T.area<minarea) {
				minarea = T.area;
				iminarea = i;
				jminarea = j;
			}

			if (T.area>maxarea) {
				maxarea = T.area;
				imaxarea = i;
				jmaxarea = j;
			}

			if (T.minangle<minangle) {
				minangle = T.minangle;
				iminangle = i;
				jminangle = j;
			}

		}
	
	printf("%d triangles\n", count);
	printf("min area = %g at (%d, %d)\n", minarea, iminarea, jminarea);
	printf("max area = %g at (%d, %d)\n", maxarea, imaxarea, jmaxarea);
	printf("min area / max area = %g\n", minarea/maxarea);
	printf("min angle = %g rad = %g deg at (%d, %d)\n",
			minangle, 180*minangle/M_PI, iminangle, jminangle);
}
#endif /* DEBUG_TRIANGLES */
