#include <stdio.h>
#include <math.h>	/* for sqrt() */
#include <string.h>	/* for strcpy() */
#include <stdlib.h>	/* for malloc() */
#include <float.h>	/* for DBL_MAX */
#include <stdarg.h>
#include <ctype.h>
#include "dynagraph.h"
#include "parse.h"

extern Expr		*symtab;
extern Plot3D		dg_plots;
extern Plot3Dattribs	dg_attribs;
extern char		*dispargs;
extern char		*prognam;

/* Evaluates the numberical value of an expression and places it in the
   supplied argument d.  Returns 1 on success, 0 on error.
*/
int evalf(Expr *e, double *d)
{
	FILE *fp;
	int ret;
	char *t;

	if (e==NULL)
		return 0;

	if ((fp=tmpfile())==NULL) {
		printf("\nunable to open temporary file tmpfile() to write\n");
		exit(1);
	}

	t = expr_eval(copy_string(e->rval));
	fprintf(fp, "print %s\n", t);
	testfree(t);

	/* call hoc to calculate */
	rewind(fp);
	ret = hoc(fp, d);
	fclose(fp);

	return ret;
}

/* Calls hoc() to numerically evaluate the data for a graph.
   Retrurns 1 on success, 0 on failure */
static int evaluate_param_plot(Plot3D *p)
{
	Plot3Dattribs *a = p->attr;
	char *xlo, *xhi, *ylo, *yhi;
	char *u, *v, *w;
	int i, j;
	int ret;
	FILE *fp;
	Expr *s;
	double d;

	if (p->attr->xgrid!=NULL && p->attr->ygrid!=NULL) {
		if (evalf(p->attr->xgrid, &d)==0) {
			printf("\terror: invalid range spec\n");
			return 0;
		}
		a->n = MAX((int)d, MIN_GRIDSIZE);
		if (evalf(p->attr->ygrid, &d)==0) {
			printf("\terror: invalid range spec\n");
			return 0;
		}
		a->m = MAX((int)d, MIN_GRIDSIZE);
	} else if (p->attr->num_points!=NULL) {
		if (evalf(p->attr->num_points, &d)==0) {
			printf("\terror: invalid argument to numpoints\n");
			return 0;
		}
		d = sqrt(fabs(d));
		a->m = a->n = MAX((int)d, MIN_GRIDSIZE);
	}

	xlo = expr_eval(copy_string(a->xrange->lo->rval));
	xhi = expr_eval(copy_string(a->xrange->hi->rval));
	ylo = expr_eval(copy_string(a->yrange->lo->rval));
	yhi = expr_eval(copy_string(a->yrange->hi->rval));
	u = expr_eval(copy_string(p->parg->u->rval));
	v = expr_eval(copy_string(p->parg->v->rval));
	w = expr_eval(copy_string(p->parg->w->rval));

	if ((fp=tmpfile())==NULL) {
		printf("\tunable to open temporary file tmpfile() to write\n");
		exit(1);
	}

	/* we use an underscore prefix for the names of temporary variavles
	   to be seen by hoc.  Since Dynagraph does not allow identifiers
	   which begin with an underscore while hoc does, this avoids
	   potential name clashes.
	 */

	fprintf(fp, "for(_i=0;_i<%d;_i++){\n", a->m);
	fprintf(fp, "\t%s=(%s)+_i*((%s)-(%s))/(%d-1)\n",
			a->yrange->id, ylo, yhi, ylo, a->m);
	fprintf(fp, "\tfor(_j=0;_j<%d; _j++){\n", a->n);
	fprintf(fp, "\t\t%s=(%s)+_j*((%s)-(%s))/(%d-1)\n",
			a->xrange->id, xlo, xhi, xlo, a->n);

	switch (p->attr->coords&~locked) {
		/* Same code for the following two; the arg vectors
		   are different, though.  (Search for the string "rotate"
		   further doww in this file to see why.) */
		case CYLINDRICAL:	/* [ u*cos(v), u*sin(v), w ] */
		case Z_CYLINDRICAL:	/* [ u*cos(v), u*sin(v), w ] */
			fprintf(fp,	"\t\tprint (%s)*cos(%s)\n"
					"\t\tprint (%s)*sin(%s)\n"
					"\t\tprint %s\n"
					"\t}\n" "}\n", u, v, u, v, w);
			break;
		case SPHERICAL:
			/* [ u*sin(w)*cos(v), u*sin(w)*sin(v), u*cos(w)] */
			fprintf(fp,	"\t\tprint (%s)*sin(%s)*cos(%s)\n"
					"\t\tprint (%s)*sin(%s)*sin(%s)\n"
					"\t\tprint (%s)*cos(%s)\n"
					"\t}\n" "}\n", u, w, v, u, w, v, u, w);
			break;
		case CARTESIAN:		/* [ u, v, w ] */
		default:
			fprintf(fp,	"\t\tprint %s\n"
					"\t\tprint %s\n"
					"\t\tprint %s\n"
					"\t}\n" "}\n", u, v, w);
			break;
	}

	testfree(xlo);
	testfree(xhi);
	testfree(ylo);
	testfree(yhi);
	testfree(u);
	testfree(v);
	testfree(w);

	p->mesh = create_mesh(a->m, a->n);
	p->normal = create_mesh(a->m, a->n);
	
	/* show the contents of tmpfile() */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"1")==0) {
		int c;
		rewind(fp);
		printf("\n=======\n");
		while ((c=getc(fp))!=EOF)
			putchar(c);
		printf("=======\n");
	}

	/* call hoc to calculate */
	rewind(fp);
	ret = hoc(fp, (double *)(p->mesh[0]));
	fclose(fp);

	/* hoc returns 0 if it is unable to evaluate an expression */
	if (ret==0) {
		printf("\tcannot evaluate expression\n");
		return 0;
	}

	compute_bounds(p);
	compute_normals(p);

	/* print out computed values */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"2")==0) {
		printf("Mesh is %d x %d:\n", a->m, a->n);
		for (i=0; i<a->m; i++) {
			for (j=0; j<a->n; j++)
				printf("[%.2g, %.2g, %.2g], ",
						p->mesh[i][j].x,
						p->mesh[i][j].y,
						p->mesh[i][j].z);
			putchar('\n');
		}
		printf("Normals:\n");
		for (i=0; i<a->m; i++) {
			for (j=0; j<a->n; j++)
				printf("[%.2g, %.2g, %.2g], ",
						p->normal[i][j].x,
						p->normal[i][j].y,
						p->normal[i][j].z);
			putchar('\n');
		}
	}

	return 1;
}

/* Numerically evaluates a function or parametric plot list.
   Returns 1 on success, 0 on failure */
static int evaluate_param_plots(Plot3D *plots)
{
	Plot3D *p;
	Plot3Dattribs *a;

	if (plots==NULL)
		return 0;

	a = plots->attr;

	for (p=plots; p!=NULL; p=p->fp) {

		/* The argument for a _function_ plot is stored in p->parg->w,
		   and the u and v slots are NULL.  In a _parametric_ plot
		   all three slots are occupied.  We convert all function
		   plots to the equivalent parametic plots. */
		
		if (p->type==FUNCTION_PLOT) {
			/* create x coordinate spec from range spec */
			p->parg->u = malloc(sizeof(Expr));
			p->parg->u->rval = copy_string(a->xrange->id);
			p->parg->u->lval = NULL;
			p->parg->u->fp = NULL;

			/* create y coordinate spec from range spec */
			p->parg->v = malloc(sizeof(Expr));
			p->parg->v->rval = copy_string(a->yrange->id);
			p->parg->v->lval = NULL;
			p->parg->v->fp = NULL;
		}

		/* rotate the coordinate assignment to conform with Maple's */
		if (p->type==FUNCTION_PLOT && ((a->coords&~locked)==CYLINDRICAL||(a->coords&~locked)==SPHERICAL)) {
			Expr *tmp;
			tmp = p->parg->w;
			p->parg->w = p->parg->v;
			p->parg->v = p->parg->u;
			p->parg->u = tmp;
		}

		if (evaluate_param_plot(p)==0)
			return 0;	/* abort on error */

		a->nplots++;

		a->xmin = MIN(a->xmin, p->xmin);
		a->xmax = MAX(a->xmax, p->xmax);
		a->ymin = MIN(a->ymin, p->ymin);
		a->ymax = MAX(a->ymax, p->ymax);
		a->zmin = MIN(a->zmin, p->zmin);
		a->zmax = MAX(a->zmax, p->zmax);
	}

	/* if z-view range is specified, use it */
	if (a->zviewmin!=DBL_MAX) {
		a->zmin = a->zviewmin;
		a->zmax = a->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (a->xviewmin!=DBL_MAX) {
		a->xmin = a->xviewmin;
		a->xmax = a->xviewmax;
		a->ymin = a->yviewmin;
		a->ymax = a->yviewmax;
	}

	compute_centroid(a);

	return 1;
}

/* Calls hoc() to numerically evaluate the data for a spacecurve.
   Retrurns 1 on success, 0 on failure */
static int evaluate_spacecurve(Plot3D *curve)
{
	char *rlo, *rhi;
	char *u, *v, *w;
	int i;
	int ret;
	FILE *fp;
	Expr *s;
	double d;

	if (evalf(curve->attr->num_points, &d)==0)
		curve->attr->n = DEFAULT_SPACECURVE_POINTS;
	else
		curve->attr->n = MAX(MIN_SPACECURVE_POINTS, (int)d);

	rlo = expr_eval(copy_string(curve->attr->xrange->lo->rval));
	rhi = expr_eval(copy_string(curve->attr->xrange->hi->rval));

	u = expr_eval(copy_string(curve->parg->u->rval));
	v = expr_eval(copy_string(curve->parg->v->rval));
	w = expr_eval(copy_string(curve->parg->w->rval));

	if ((fp=tmpfile())==NULL) {
		printf("\tunable to open temporary file tmpfile() to write\n");
		exit(1);
	}

	/* we use an underscore prefix for the names of temporary variavles
	   to be seen by hoc.  Since Dynagraph does not allow identifiers
	   which begin with an underscore while hoc does, this avoids
	   potential name clashes.
	 */

	fprintf(fp, "for(_i=0;_i<%d;_i++){\n", curve->attr->n);
	fprintf(fp, "\t%s=(%s)+_i*((%s)-(%s))/(%d-1)\n",
			curve->attr->xrange->id, rlo, rhi, rlo, curve->attr->n);
	fprintf(fp,	"\tprint %s\n"
			"\tprint %s\n"
			"\tprint %s\n"
			"}\n", u, v, w);

	testfree(rlo);
	testfree(rhi);
	testfree(u);
	testfree(v);
	testfree(w);

	curve->curvepoints = create_curvepoints(curve->attr->n);
	
	/* show the contents of tmpfile() */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"1")==0) {
		int c;
		rewind(fp);
		printf("\n=======\n");
		while ((c=getc(fp))!=EOF)
			putchar(c);
		printf("=======\n");
	}

	/* call hoc to calculate */
	rewind(fp);
	ret = hoc(fp, (double *)(curve->curvepoints));
	fclose(fp);

	/* hoc returns 0 if it is unable to evaluate an expression */
	if (ret==0) {
		printf("\terror: cannot evaluate expression\n");
		return 0;
	}

	compute_curvebounds(curve);

	/* print out computed values */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"2")==0) {
		for (i=0; i<curve->attr->n; i++) {
			printf("[%.2g, %.2g, %.2g], ",
					curve->curvepoints[i].x,
					curve->curvepoints[i].y,
					curve->curvepoints[i].z);
			putchar('\n');
		}
	}

	return 1;
}

/* Numerically evaluates a list of spacecurves.  Returns 1 on success,
   0 on failure */
static int evaluate_spacecurves(Plot3D *curves)
{
	Plot3D *p;
	Plot3Dattribs *a;

	if (curves==NULL)
		return 0;

	a = curves->attr;

	for (p=curves; p!=NULL; p=p->fp) {

		if (evaluate_spacecurve(p)==0)
			return 0;	/* abort on error */

		a->nplots++;

		a->xmin = MIN(a->xmin, p->xmin);
		a->xmax = MAX(a->xmax, p->xmax);
		a->ymin = MIN(a->ymin, p->ymin);
		a->ymax = MAX(a->ymax, p->ymax);
		a->zmin = MIN(a->zmin, p->zmin);
		a->zmax = MAX(a->zmax, p->zmax);
	}

	/* if z-view range is specified, use it */
	if (a->zviewmin!=DBL_MAX) {
		a->zmin = a->zviewmin;
		a->zmax = a->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (a->xviewmin!=DBL_MAX) {
		a->xmin = a->xviewmin;
		a->xmax = a->xviewmax;
		a->ymin = a->yviewmin;
		a->ymax = a->yviewmax;
	}

	compute_centroid(a);

	return 1;
}

/* computes the cross-product c of two vectors a and b */
static void cross_product(Point a, Point b, Point *c)
{
	c->x = a.y*b.z - a.z*b.y;
	c->y = a.z*b.x - a.x*b.z;
	c->z = a.x*b.y - a.y*b.x;
}

/* computes dot-product */
static double dot_product(Point a, Point b)
{
	return a.x*b.x + a.y*b.y + a.z*b.z;
}

/* Numerically evaluates polygon vertices.  Returns 1 on success,
   0 on failure
*/
static int evaluate_polygonplot(Plot3D *polygon)
{
	Plot3Dattribs *attr;
	PolygonVertex *vertex;
	Point o, a, b;
	int n;		/* number of vertices */
	double len;

	if (polygon==NULL)
		return 0;

	attr = polygon->attr;

	for (n=0, vertex=polygon->polygon_vertices; vertex!=NULL;
			n++, vertex=vertex->fp) {
		Expr3 *e3 = vertex->v;
		evalf(e3->u, &vertex->numeric.x);
		evalf(e3->v, &vertex->numeric.y);
		evalf(e3->w, &vertex->numeric.z);

		polygon->xmin = MIN(vertex->numeric.x, polygon->xmin);
		polygon->xmax = MAX(vertex->numeric.x, polygon->xmax);
		polygon->ymin = MIN(vertex->numeric.y, polygon->ymin);
		polygon->ymax = MAX(vertex->numeric.y, polygon->ymax);
		polygon->zmin = MIN(vertex->numeric.z, polygon->zmin);
		polygon->zmax = MAX(vertex->numeric.z, polygon->zmax);
	}

	if (n<3)
		return 0;	/* need at least 3 vertices */

	attr->xmin = polygon->xmin;
	attr->xmax = polygon->xmax;
	attr->ymin = polygon->ymin;
	attr->ymax = polygon->ymax;
	attr->zmin = polygon->zmin;
	attr->zmax = polygon->zmax;

	/* if z-view range is specified, use it */
	if (attr->zviewmin!=DBL_MAX) {
		attr->zmin = attr->zviewmin;
		attr->zmax = attr->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (attr->xviewmin!=DBL_MAX) {
		attr->xmin = attr->xviewmin;
		attr->xmax = attr->xviewmax;
		attr->ymin = attr->yviewmin;
		attr->ymax = attr->yviewmax;
	}

	compute_centroid(attr);

	/* compute normal */
	vertex = polygon->polygon_vertices;
	o.x = vertex->numeric.x;
	o.y = vertex->numeric.y;
	o.z = vertex->numeric.z;

	vertex = vertex->fp;
	a.x = vertex->numeric.x;
	a.y = vertex->numeric.y;
	a.z = vertex->numeric.z;

	vertex = vertex->fp;
	b.x = vertex->numeric.x;
	b.y = vertex->numeric.y;
	b.z = vertex->numeric.z;

	a.x -= o.x;
	a.y -= o.y;
	a.z -= o.z;

	b.x -= o.x;
	b.y -= o.y;
	b.z -= o.z;

	cross_product(a,b,&polygon->polygon_normal);

	len = dot_product(polygon->polygon_normal, polygon->polygon_normal);
	len = sqrt(len);

	polygon->polygon_normal.x /= len;
	polygon->polygon_normal.y /= len;
	polygon->polygon_normal.z /= len;

	return 1;
}

/* Calls hoc() to numerically evaluate the data for a tubeplot.
   Retrurns 1 on success, 0 on failure */
static int evaluate_tubeplot(Plot3D *tube)
{
	Plot3Dattribs *a = tube->attr;
	char *rlo, *rhi;
	char *u, *v, *w;
	int i, j;
	int ret;
	FILE *fp;
	Expr *s;
	double d;

	if (evalf(tube->attr->num_points, &d)==0)
		tube->attr->n = DEFAULT_SPACECURVE_POINTS;
	else
		tube->attr->n = MAX(MIN_SPACECURVE_POINTS, (int)d);

	if (evalf(tube->attr->tube_points, &d)==0)
		tube->attr->m = DEFAULT_TUBE_POINTS;
	else
		tube->attr->m = MAX(MIN_TUBE_POINTS, (int)d);

	if (evalf(tube->attr->tube_radius, &d)==0)
		tube->attr->tuberadius = DEFAULT_TUBE_RADIUS;
	else
		tube->attr->tuberadius = MAX(0, d);

	rlo = expr_eval(copy_string(tube->attr->xrange->lo->rval));
	rhi = expr_eval(copy_string(tube->attr->xrange->hi->rval));

	u = expr_eval(copy_string(tube->parg->u->rval));
	v = expr_eval(copy_string(tube->parg->v->rval));
	w = expr_eval(copy_string(tube->parg->w->rval));

	if ((fp=tmpfile())==NULL) {
		printf("\tunable to open temporary file tmpfile() to write\n");
		exit(1);
	}

	/* we use an underscore prefix for the names of temporary variavles
	   to be seen by hoc.  Since Dynagraph does not allow identifiers
	   which begin with an underscore while hoc does, this avoids
	   potential name clashes.
	 */

	fprintf(fp, "for(_i=0;_i<%d;_i++){\n", tube->attr->n);
	fprintf(fp, "\t%s=(%s)+_i*((%s)-(%s))/(%d-1)\n",
			tube->attr->xrange->id, rlo, rhi, rlo, tube->attr->n);
	fprintf(fp,	"\tprint %s\n"
			"\tprint %s\n"
			"\tprint %s\n"
			"}\n", u, v, w);

	testfree(rlo);
	testfree(rhi);
	testfree(u);
	testfree(v);
	testfree(w);

	tube->curvepoints = create_curvepoints(tube->attr->n);
	
	/* show the contents of tmpfile() */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"1")==0) {
		int c;
		rewind(fp);
		printf("\n=======\n");
		while ((c=getc(fp))!=EOF)
			putchar(c);
		printf("=======\n");
	}

	/* call hoc to calculate */
	rewind(fp);
	ret = hoc(fp, (double *)(tube->curvepoints));
	fclose(fp);

	/* hoc returns 0 if it is unable to evaluate an expression */
	if (ret==0) {
		printf("\terror: cannot evaluate expression\n");
		return 0;
	}

	tube->mesh = maketube(tube->curvepoints, a->n, a->m, tube->attr->tuberadius);
	if (tube->mesh==NULL)
		return 0;
	tube->normal = create_mesh(a->m, a->n);

	compute_bounds(tube);
	compute_normals(tube);

	/* print out computed values */
	if ((s=getsym("verbose"))!=NULL && strcmp(s->rval,"2")==0) {
		for (i=0; i<a->m; i++) {
			for (j=0; j<a->n; j++)
				printf("[%.2g, %.2g, %.2g], ",
						tube->mesh[i][j].x,
						tube->mesh[i][j].y,
						tube->mesh[i][j].z);
			putchar('\n');
		}
	}

	return 1;
}

/* Numerically evaluates a list of tubeplots.  Returns 1 on success,
   0 on failure */
static int evaluate_tubeplots(Plot3D *tubes)
{
	Plot3D *p;
	Plot3Dattribs *a;

	if (tubes==NULL)
		return 0;

	a = tubes->attr;

	for (p=tubes; p!=NULL; p=p->fp) {

		if (evaluate_tubeplot(p)==0)
			return 0;	/* abort on error */

		a->nplots++;

		a->xmin = MIN(a->xmin, p->xmin);
		a->xmax = MAX(a->xmax, p->xmax);
		a->ymin = MIN(a->ymin, p->ymin);
		a->ymax = MAX(a->ymax, p->ymax);
		a->zmin = MIN(a->zmin, p->zmin);
		a->zmax = MAX(a->zmax, p->zmax);
	}

	/* if z-view range is specified, use it */
	if (a->zviewmin!=DBL_MAX) {
		a->zmin = a->zviewmin;
		a->zmax = a->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (a->xviewmin!=DBL_MAX) {
		a->xmin = a->xviewmin;
		a->xmax = a->xviewmax;
		a->ymin = a->yviewmin;
		a->ymax = a->yviewmax;
	}

	compute_centroid(a);

	return 1;
}

static int evaluate_textplot(Plot3D *p)
{
	Plot3Dattribs *a = p->attr;
	TextObj *t = p->textobj;
	double d;

	if (p==NULL)
		return 0;

	a->nplots = 1;

	while (t!=NULL) {
		if (evalf(t->xspec, &d)==0) {
			printf("\terror: unable to evaluate text coordinate %s\n",
					t->xspec->rval);
			return 0;
		} else
			t->x = d;
		if (evalf(t->yspec, &d)==0) {
			printf("\terror: unable to evaluate text coordinate %s\n",
					t->yspec->rval);
			return 0;
		} else
			t->y = d;
		if (evalf(t->zspec, &d)==0) {
			printf("\terror: unable to evaluate text coordinate %s\n",
					t->zspec->rval);
			return 0;
		} else
			t->z = d;

		/* this is not exactly right, because it treats text as
		   point rather than as something which has width and height */
		a->xmin = MIN(a->xmin, t->x);
		a->xmax = MAX(a->xmax, t->x);
		a->ymin = MIN(a->ymin, t->y);
		a->ymax = MAX(a->ymax, t->y);
		a->zmin = MIN(a->zmin, t->z);
		a->zmax = MAX(a->zmax, t->z);

		t = t->fp;
	}

	/* if z-view range is specified, use it */
	if (a->zviewmin!=DBL_MAX) {
		a->zmin = a->zviewmin;
		a->zmax = a->zviewmax;
	}

	/* if x-view and y-view ranges are specified, use them */
	if (a->xviewmin!=DBL_MAX) {
		a->xmin = a->xviewmin;
		a->xmax = a->xviewmax;
		a->ymin = a->yviewmin;
		a->ymax = a->yviewmax;
	}

	compute_centroid(a);

	return 1;
}

int evaluate_plots(Plot3D *p)
{
	double d;

	if (evalf(p->attr->x_viewmin, &d)!=0)
		p->attr->xviewmin = d;
	if (evalf(p->attr->x_viewmax, &d)!=0)
		p->attr->xviewmax = d;
	if (evalf(p->attr->y_viewmin, &d)!=0)
		p->attr->yviewmin = d;
	if (evalf(p->attr->y_viewmax, &d)!=0)
		p->attr->yviewmax = d;
	if (evalf(p->attr->z_viewmin, &d)!=0)
		p->attr->zviewmin = d;
	if (evalf(p->attr->z_viewmax, &d)!=0)
		p->attr->zviewmax = d;

	if (evalf(p->attr->view_longitude, &d)!=0)
		p->attr->longitude = d;
	if (evalf(p->attr->view_colatitude, &d)!=0)
		p->attr->colatitude = d;

	if (evalf(p->attr->num_contours, &d)!=0)
		p->attr->contours = d;

	if (p->type == FUNCTION_PLOT || p->type == PARAM_PLOT)
		return evaluate_param_plots(p);
	else if (p->type == CURVE_PLOT)
		return evaluate_spacecurves(p);
	else if (p->type == TUBE_PLOT)
		return evaluate_tubeplots(p);
	else if (p->type == TEXT_PLOT)
		return evaluate_textplot(p);
	else if (p->type == POLYGON_PLOT)
		return evaluate_polygonplot(p);

	printf("%s internal error! %s line %d: should not have arrived here\n",
			prognam, __FILE__, __LINE__);
	exit(0);
}
