#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "dynagraph.h"

#define TINY		1e-4

static Point	*normal;
static Point	*binormal;
static double	*segment_length;
extern char	*prognam;

/* computes dot-product */
static double dot_product(Point a, Point b)
{
	return a.x*b.x + a.y*b.y + a.z*b.z;
}

/* computes the cross-product c of two vectors a and b */
static void cross_product(Point a, Point b, Point *c)
{
	c->x = a.y*b.z - a.z*b.y;
	c->y = a.z*b.x - a.x*b.z;
	c->z = a.x*b.y - a.y*b.x;
}

/* Removes point j from the array, shifting remaining points downwards. */
static void remove_point(Point *p, int n, int j)
{
	int k;

	for (k=j; k<n-1; k++) {
		p[k].x = p[k+1].x;
		p[k].y = p[k+1].y;
		p[k].z = p[k+1].z;
	}
}

/* Removes singular points from array, shifting the remaining points downwards.
   Returns the length of the revised array. */
static int remove_singularities(Point *p, int n)
{
	int j = 0;

	while (j<n-1)
		if (IS_SINGULAR(p[j]))
			remove_point(p, n--, j);
		else
			j++;
	return n;
}

/* 1. Receives an array of length n of points.  When two adjacent points
      coincide, removes the second of the two, shifting the remaining
      points downwards.
   
   2. Computes and saves in segment_length[] the (non-zero) distances
      between consecutive points of the revised array.

   3. Returns the length of the revised array.
*/
static int compute_link_lengths(Point *p, int n)
{
	double x, y, z, len;
	int j, k;

	j = k = 0;
	while (j<n-1) {
		x = p[j+1].x - p[j].x;
		y = p[j+1].y - p[j].y;
		z = p[j+1].z - p[j].z;
		if ((len=sqrt(x*x+y*y+z*z))<TINY)
			remove_point(p, n--, j+1);
		else {
			segment_length[k++] = len;
			j++;
		}
	}
	return n;
}

/* normalizes a vector, returns the orignial length */
double normalize(Point *p)
{
	double len = sqrt(p->x*p->x+p->y*p->y+p->z*p->z);
	p->x /= len;
	p->y /= len;
	p->z /= len;
	return len;
}

/* computes normal and binornal vectors at one end of the curve */
static void compute_initial_normals(Point *p)
{
	double u1, v1, w1;
	double u2, v2, w2;
	double x, y, z;
	Point t;

	/* unit vector from p[-1] to p[0] */
	u1 = (p[0].x - p[-1].x)/segment_length[-1];
	v1 = (p[0].y - p[-1].y)/segment_length[-1];
	w1 = (p[0].z - p[-1].z)/segment_length[-1];

	/* unit vector from p[0] to p[1] */
	u2 = (p[1].x - p[0].x)/segment_length[0];
	v2 = (p[1].y - p[0].y)/segment_length[0];
	w2 = (p[1].z - p[0].z)/segment_length[0];

	/* sum of u and v gives the "tangent vector" */
	t.x = u1 + u2;
	t.y = v1 + v2;
	t.z = w1 + w2;

	/* this will happen if p[-1] == p[+1] */
	if (normalize(&t)<TINY)
		printf("\twarning: degenerate point on curve!\n");

	x = y = z = 1.0;
	if (fabs(t.x)>=fabs(t.y) && fabs(t.x)>=fabs(t.z))
		x = -(t.y+t.z)/t.x;
	else if (fabs(t.y)>=fabs(t.x) && fabs(t.y)>=fabs(t.z))
		y = -(t.x+t.z)/t.y;
	else
		z = -(t.x+t.y)/t.z;

	binormal[0].x = x;
	binormal[0].y = y;
	binormal[0].z = z;
	normalize(&binormal[0]);

	cross_product(binormal[0], t, &normal[0]);
}

/* computes normal and binormal vectors at all points of the curve
   assuming the these vectors are known at one end.
*/
static void compute_curve_normals(Point *p, int n) {

	double ux, uy, uz, vx, vy, vz;
	Point t;
	int j;

	for (j=1; j<n; j++) {
		/* unit vector from p[j-1] to p[j] */
		ux = (p[j].x - p[j-1].x)/segment_length[j-1];
		uy = (p[j].y - p[j-1].y)/segment_length[j-1];
		uz = (p[j].z - p[j-1].z)/segment_length[j-1];

		/* unit vector from p[j] to p[j+1] */
		vx = (p[j+1].x - p[j].x)/segment_length[j];
		vy = (p[j+1].y - p[j].y)/segment_length[j];
		vz = (p[j+1].z - p[j].z)/segment_length[j];

		/* sum of u and v gives the "tangent vector" */
		t.x = vx + ux;
		t.y = vy + uy;
		t.z = vz + uz;

		/* this will happen if p[j-1] == p[j+1] */
		if (normalize(&t)<TINY)
			printf("\twarning: degenerate point on curve!\n");

		cross_product(binormal[j-1], t, &normal[j]);
		if (normalize(&normal[j])<TINY) {
			cross_product(t, normal[j-1], &binormal[j]);
			normalize(&binormal[j]);
			cross_product(binormal[j], t, &normal[j]);
		} else
			cross_product(t, normal[j], &binormal[j]);
	}
}

/* Computes and inserts in mesh[][] the grid points for circles around nodes */
static void put_rings(Point **mesh, int m, int n, Point *p, double r, short closed)
{
	double a, adj;
	int i, j;

	/* if curve is closed, compute the angle by which the coordinate
	   axes get twisted when we go one round around, then compensate
	   so that the tuble closes nicely on itself. */
	adj = closed ? atan2(
			dot_product(normal[n-1], binormal[0]),
			dot_product(normal[n-1], normal[0]))
		: 0;

	for (j=0; j<n; j++) {
		for (i=0; i<m; i++) {
			a = i*2*M_PI/(m-1) - j*adj/(n-1);
			mesh[i][j].x = p[j].x
				+ r * normal[j].x * cos(a)
				+ r * binormal[j].x * sin(a);
			mesh[i][j].y = p[j].y
				+ r * normal[j].y * cos(a)
				+ r * binormal[j].y * sin(a);
			mesh[i][j].z = p[j].z
				+ r * normal[j].z * cos(a)
				+ r * binormal[j].z * sin(a);
		}
	}
}

Point **maketube(Point *p, int n, int m, double r)
{
	Point **mesh;
	short closed;
	int i, j;
	double *tmp1;
	Point *tmp2, *curve;

	n = remove_singularities(p, n);

	/* segment_length[] will hold the (n-1) lengths of the segments
	   of the given curve.  We set aside an extra storage at the location
	   segment_length[-1] and another at segment_length[n] which will
	   hold values for continuations of the curve. */
	tmp1 = malloc((n+1)*sizeof(double));
	segment_length = tmp1+1;

	n = compute_link_lengths(p, n);

	if (n<3) {
		printf("\nerror: too few usable data points on curve\n");
		free(tmp1);
		return NULL;
	}

	if (fabs(p[0].x - p[n-1].x)<TINY && fabs(p[0].y - p[n-1].y)<TINY &&
			fabs(p[0].z - p[n-1].z)<TINY)
		closed = 1;
	else
		closed = 0;

	/* make a copy of the received curve into a larger array by
	   adding extra slots at the beginning and at the end
	   to help handle end conditions */
	tmp2 = malloc((n+2)*sizeof(Point));
	curve = tmp2+1;

	for (j=0; j<n; j++)
		curve[j] = p[j];

	if (closed) {
		curve[-1].x = p[n-2].x;
		curve[-1].y = p[n-2].y;
		curve[-1].z = p[n-2].z;
		segment_length[-1] = segment_length[n-2];
		curve[n].x  = p[1].x;
		curve[n].y  = p[1].y;
		curve[n].z  = p[1].z;
		segment_length[n-1] = segment_length[0];
	} else {
		curve[-1].x = p[0].x - (p[1].x-p[0].x);
		curve[-1].y = p[0].y - (p[1].y-p[0].y);
		curve[-1].z = p[0].z - (p[1].z-p[0].z);
		segment_length[-1] = segment_length[0];
		curve[n].x  = p[n-1].x + (p[n-1].x - p[n-2].x);
		curve[n].y  = p[n-1].y + (p[n-1].y - p[n-2].y);
		curve[n].z  = p[n-1].z + (p[n-1].z - p[n-2].z);
		segment_length[n-1] = segment_length[n-2];
	}

	normal = malloc(n*sizeof(Point));
	binormal = malloc(n*sizeof(Point));
	compute_initial_normals(curve);
	compute_curve_normals(curve, n);

	mesh = create_mesh(m, n);

	put_rings(mesh, m, n, p, r, closed);

	/* if a closed curve, make sure the first and last circles coincide */
	if (closed) {
		for (i=0; i<m; i++) {
			mesh[i][0].x = mesh[i][n-1].x;
			mesh[i][0].y = mesh[i][n-1].y;
			mesh[i][0].z = mesh[i][n-1].z;
		}
	}

	free(tmp1);
	free(tmp2);
	free(normal);
	free(binormal);

	return mesh;
}
